from typing import Any, Optional, Union, cast

import fmf.utils

import tmt.base
import tmt.log
import tmt.result
import tmt.steps
import tmt.steps.discover.fmf
import tmt.steps.execute
import tmt.steps.provision
import tmt.utils
from tmt.container import container, field, key_to_option
from tmt.steps.discover import Discover, DiscoverPlugin, DiscoverStepData
from tmt.steps.discover.fmf import DiscoverFmf, DiscoverFmfStepData, normalize_ref
from tmt.steps.execute import ExecutePlugin
from tmt.steps.execute.internal import ExecuteInternal, ExecuteInternalData
from tmt.steps.prepare import PreparePlugin
from tmt.steps.prepare.install import PrepareInstallData
from tmt.utils import Environment, EnvVarValue, Path

STATUS_VARIABLE = 'IN_PLACE_UPGRADE'
BEFORE_UPGRADE_PREFIX = 'old'
DURING_UPGRADE_PREFIX = 'upgrade'
AFTER_UPGRADE_PREFIX = 'new'
UPGRADE_DIRECTORY = 'upgrade'

PROPAGATE_TO_DISCOVER_KEYS = ['url', 'ref', 'filter', 'test', 'exclude', 'upgrade_path']


@container
class ExecuteUpgradeData(ExecuteInternalData):
    url: Optional[str] = field(
        default=cast(Optional[str], None),
        option=('-u', '--url'),
        metavar='REPOSITORY',
        help='URL of the git repository with upgrade tasks.',
    )
    upgrade_path: Optional[str] = field(
        default=cast(Optional[str], None),
        option=('-p', '--upgrade-path'),
        metavar='PLAN_NAME',
        help='Upgrade path corresponding to a plan name in the repository with upgrade tasks.',
    )

    # "Inherit" from tmt.steps.discover.fmf.DiscoverFmfStepData
    ref: Optional[str] = field(
        default=cast(Optional[str], None),
        option=('-r', '--ref'),
        metavar='REVISION',
        help='Branch, tag or commit specifying the git revision.',
        normalize=normalize_ref,
    )
    test: list[str] = field(
        default_factory=list,
        option=('-t', '--test'),
        metavar='NAMES',
        multiple=True,
        help='Select tests by name.',
        normalize=tmt.utils.normalize_string_list,
    )
    filter: list[str] = field(
        default_factory=list,
        option=('-F', '--filter'),
        metavar='FILTERS',
        multiple=True,
        help='Include only tests matching the filter.',
        normalize=tmt.utils.normalize_string_list,
    )
    exclude: list[str] = field(
        default_factory=list,
        option=('-x', '--exclude'),
        metavar='REGEXP',
        multiple=True,
        help="Exclude a regular expression from search result.",
        normalize=tmt.utils.normalize_string_list,
    )


@tmt.steps.provides_method('upgrade')
class ExecuteUpgrade(ExecuteInternal):
    """
    Perform system upgrade during testing.

    In order to enable developing tests for upgrade testing, we need to provide
    a way how to execute these tests easily. This does not cover unit tests for
    individual actors but rather system tests which verify
    the whole upgrade story.

    The upgrade executor runs the discovered tests (using the internal
    executor), then performs a set of upgrade tasks from a remote
    repository, and finally, re-runs the tests on the upgraded guest.

    The ``IN_PLACE_UPGRADE`` environment variable is set during the test
    execution to differentiate between the stages of the test. It is set
    to ``old`` during the first execution and ``new`` during the second
    execution. Test names are prefixed with this value to make the names
    unique. Based on this variable, the test can perform appropriate actions.

    * ``old``: setup, test
    * ``new``: test, cleanup
    * ``without``: setup, test, cleanup

    The upgrade tasks performing the actual system upgrade are taken
    from a remote repository (specified by the ``url`` key) based on an upgrade
    path (e.g. ``fedora35to36``) or other filters (e.g. specified by the
    ``filter`` key). If both ``upgrade-path`` and extra filters are specified,
    the discover keys in the remote upgrade path plan are overridden by the
    filters specified in the local plan.

    The upgrade path must correspond to a plan name in the
    remote repository whose discover step selects tests (upgrade tasks)
    performing the upgrade. Currently, selection of upgrade tasks in the remote
    repository can be done using both fmf and shell discover method.
    If the ``url`` is not provided, upgrade path and upgrade tasks are taken from
    the current repository. The supported keys in discover are:

    * ``ref``
    * ``filter``
    * ``exclude``
    * ``tests``
    * ``test``

    The environment variables defined in the remote upgrade path plan are
    passed to the upgrade tasks when they are executed. An example of an
    upgrade path plan (in the remote repository):

    .. code-block:: yaml

        discover: # Selects appropriate upgrade tasks (L1 tests)
            how: fmf
            filter: "tag:fedora"
        environment: # This is passed to upgrade tasks
            SOURCE: 35
            TARGET: 36
        execute:
            how: tmt

    If no upgrade path is specified in the plan, the tests (upgrade tasks)
    are selected based on the configuration of the upgrade plugin
    (e.g. based on the filter in its configuration).

    If these two possible ways of specifying upgrade tasks are combined,
    the remote discover plan is used but its options are overridden
    with the values specified locally.

    The same options and config keys and values can be used as in the
    internal executor.

    Minimal execute config example with an upgrade path:

    .. code-block:: yaml

        execute:
            how: upgrade
            url: https://github.com/teemtee/upgrade
            upgrade-path: /paths/fedora35to36

    Execute config example without an upgrade path:

    .. code-block:: yaml

        execute:
            how: upgrade
            url: https://github.com/teemtee/upgrade
            filter: "tag:fedora"

    .. code-block:: yaml

        # A simple beakerlib test using the $IN_PLACE_UPGRADE variable
        . /usr/share/beakerlib/beakerlib.sh || exit 1

        VENV_PATH=/var/tmp/venv_test

        rlJournalStart
            # Perform the setup only for the old distro
            if [[ "$IN_PLACE_UPGRADE" !=  "new" ]]; then
                rlPhaseStartSetup
                    rlRun "python3.9 -m venv $VENV_PATH"
                    rlRun "$VENV_PATH/bin/pip install pyjokes"
                rlPhaseEnd
            fi

            # Execute the test for both old & new distro
            rlPhaseStartTest
                rlAsssertExists "$VENV_PATH/bin/pyjoke"
                rlRun "$VENV_PATH/bin/pyjoke"
            rlPhaseEnd

            # Skip the cleanup phase when on the old distro
            if [[ "$IN_PLACE_UPGRADE" !=  "old" ]]; then
                rlPhaseStartCleanup
                    rlRun "rm -rf $VENV_PATH"
                rlPhaseEnd
            fi
        rlJournalEnd
    """

    _data_class = ExecuteUpgradeData
    data: ExecuteUpgradeData

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._discover_upgrade: Optional[DiscoverFmf] = None

    @property  # type:ignore[override]
    def discover(self) -> Union[Discover, DiscoverFmf]:
        """
        Return discover plugin instance
        """

        # If we are in the second phase (upgrade), take tests from our fake
        # discover plugin.
        if self._discover_upgrade:
            return self._discover_upgrade
        return self.step.plan.discover

    @discover.setter
    def discover(self, plugin: Optional[DiscoverPlugin[DiscoverStepData]]) -> None:
        self._discover = plugin

    def go(
        self,
        *,
        guest: 'tmt.steps.provision.Guest',
        environment: Optional[tmt.utils.Environment] = None,
        logger: tmt.log.Logger,
    ) -> None:
        """
        Execute available tests
        """

        # Inform about the how, skip the actual execution
        ExecutePlugin.go(self, guest=guest, environment=environment, logger=logger)

        self.url = self.get('url')
        self.upgrade_path = self.get('upgrade-path')
        for key in self._keys:
            value = self.get(key)
            if value:
                if key == "test":
                    self.info('test', fmf.utils.listed(value), 'green')
                else:
                    self.info(key, value, color='green')

        # Nothing to do in dry mode
        if self.is_dry_run:
            self._results = []
            return

        self.verbose('upgrade', 'run tests on the old system', color='blue', shift=1)
        self._run_test_phase(guest, BEFORE_UPGRADE_PREFIX, logger)
        self.verbose('upgrade', 'perform the system upgrade', color='blue', shift=1)
        self._perform_upgrade(guest, logger)
        self.verbose('upgrade', 'run tests on the new system', color='blue', shift=1)
        self._run_test_phase(guest, AFTER_UPGRADE_PREFIX, logger)

    def _get_plan(self, upgrades_repo: Path) -> tmt.base.Plan:
        """
        Get plan based on upgrade path
        """

        tree = tmt.base.Tree(logger=self._logger, path=upgrades_repo)
        try:
            # We do not want to consider plan -n provided on the command line
            # in the remote repo for finding upgrade path.
            tmt.base.Plan.ignore_class_options = True
            plans = tree.plans(names=[self.upgrade_path])
        finally:
            tmt.base.Plan.ignore_class_options = False

        if len(plans) == 0:
            raise tmt.utils.ExecuteError(
                f"No matching upgrade path found for '{self.upgrade_path}'."
            )
        if len(plans) > 1:
            names = [plan.name for plan in plans]
            raise tmt.utils.ExecuteError(
                f"Ambiguous upgrade path reference, found plans {fmf.utils.listed(names)}."
            )
        return plans[0]

    def _fetch_upgrade_tasks(self) -> None:
        """
        Fetch upgrade tasks using DiscoverFmf
        """

        data = DiscoverFmfStepData(
            name='upgrade-discover',
            how='fmf',
            # url=self.data.url,
            **{key: getattr(self.data, key) for key in PROPAGATE_TO_DISCOVER_KEYS},
        )

        self._discover_upgrade = DiscoverFmf(logger=self._logger, step=self.step, data=data)
        self._run_discover_upgrade()

    def _run_discover_upgrade(self) -> None:
        """
        Silently run discover upgrade
        """

        # Make it quiet, we do not want any output from discover
        assert self._discover_upgrade is not None

        # Discover normally uses also options from global Test class
        # (e.g. test -n foo). Ignore this when selecting upgrade tasks.
        tmt.base.Test.ignore_class_options = True

        cli_invocation = self._discover_upgrade.cli_invocation
        if cli_invocation:
            quiet, cli_invocation.options['quiet'] = cli_invocation.options['quiet'], True

        try:
            self._discover_upgrade.wake()
            self._discover_upgrade.go()

        finally:
            tmt.base.Test.ignore_class_options = False

            if cli_invocation:
                cli_invocation.options['quiet'] = quiet

    def _install_dependencies(
        self,
        guest: tmt.steps.provision.Guest,
        dependencies: list[tmt.base.DependencySimple],
        recommends: bool = False,
    ) -> None:
        """
        Install packages required/recommended for upgrade
        """

        phase_name = 'recommended' if recommends else 'required'
        data = PrepareInstallData(
            how='install',
            name=f'{phase_name}-packages-upgrade',
            summary=f'Install packages {phase_name} by the upgrade',
            package=tmt.utils.uniq(dependencies),
            missing='skip' if recommends else 'fail',
        )

        PreparePlugin.delegate(self.step, data=data).go(  # type:ignore[attr-defined]
            guest=guest, logger=self._logger
        )

    def _prepare_remote_discover_data(self, plan: tmt.base.Plan) -> tmt.steps._RawStepData:
        """
        Merge remote discover data with the local filters
        """

        if len(plan.discover.data) > 1:
            raise tmt.utils.ExecuteError("Multiple discover configs are not supported.")

        data = plan.discover.data[0]

        remote_raw_data: tmt.steps._RawStepData = {
            # Force name
            'name': 'upgrade-discover-remote',
            'how': 'fmf',
        }
        remote_raw_data.update(
            cast(
                tmt.steps._RawStepData,
                {
                    key_to_option(key): value
                    for key, value in data.items()
                    if key in PROPAGATE_TO_DISCOVER_KEYS
                },
            )
        )

        # Local values have priority, override
        for key in self._keys:
            value = self.get(key)
            if key in PROPAGATE_TO_DISCOVER_KEYS and value:
                remote_raw_data[key] = value  # type:ignore[literal-required]

        return remote_raw_data

    def _perform_upgrade(self, guest: tmt.steps.provision.Guest, logger: tmt.log.Logger) -> None:
        """
        Perform a system upgrade
        """

        original_discover_phase = self.discover_phase

        try:
            self._fetch_upgrade_tasks()
            extra_environment = None
            assert self._discover_upgrade is not None
            if self.upgrade_path:
                # Create a fake discover from the data in the upgrade path
                plan = self._get_plan(self._discover_upgrade.testdir)
                data = self._prepare_remote_discover_data(plan)
                # Unset `url` because we don't want discover step to perform clone. Instead,
                # we want it to reuse existing, already cloned path.
                # ignore[typeddict-unknown-key]: data is _RwStepData, we do not have more detailed
                # type for raw step data of internal/upgrade plugins, it would be pretty verbose.
                data['url'] = None  # type: ignore[typeddict-unknown-key]
                data['path'] = self._discover_upgrade.testdir  # type:ignore[typeddict-unknown-key]
                # FIXME: cast() - https://github.com/teemtee/tmt/issues/1599
                self._discover_upgrade = cast(
                    DiscoverFmf, DiscoverPlugin.delegate(self.step, raw_data=data)
                )
                self._run_discover_upgrade()
                # Pass in the path-specific env variables
                extra_environment = plan.environment

            required_packages: list[tmt.base.DependencySimple] = []
            recommended_packages: list[tmt.base.DependencySimple] = []
            for test_origin in self._discover_upgrade.tests(enabled=True):
                test = test_origin.test

                test.name = f'/{DURING_UPGRADE_PREFIX}/{test.name.lstrip("/")}'

                # Gathering dependencies for upgrade tasks
                required_packages += tmt.base.assert_simple_dependencies(
                    test.require,
                    'After beakerlib processing, tests may have only simple requirements',
                    self._logger,
                )

                recommended_packages += tmt.base.assert_simple_dependencies(
                    test.recommend,
                    'After beakerlib processing, tests may have only simple requirements',
                    self._logger,
                )

                required_packages += test.test_framework.get_requirements(test, self._logger)

                for check in test.check:
                    required_packages += check.plugin.essential_requires(guest, test, self._logger)

            self._install_dependencies(guest, required_packages)
            self._install_dependencies(guest, recommended_packages, recommends=True)
            self.discover_phase = self._discover_upgrade.name
            self._run_tests(guest=guest, extra_environment=extra_environment, logger=logger)
        finally:
            self._discover_upgrade = None
            self.discover_phase = original_discover_phase

    def _run_test_phase(
        self, guest: tmt.steps.provision.Guest, prefix: str, logger: tmt.log.Logger
    ) -> None:
        """
        Execute a single test phase on the guest

        Tests names are prefixed with the prefix argument in order to make
        their names unique so that the results are distinguishable.
        The prefix is also set as IN_PLACE_UPGRADE environment variable.
        """

        names_backup = []
        for test_origin in self.discover.tests(enabled=True):
            names_backup.append(test_origin.test.name)
            test_origin.test.name = f'/{prefix}/{test_origin.test.name.lstrip("/")}'

        self._run_tests(
            guest=guest,
            extra_environment=Environment({STATUS_VARIABLE: EnvVarValue(prefix)}),
            logger=logger,
        )

        self._remove_old_results(prefix)

        for i, test_origin in enumerate(self.discover.tests(enabled=True)):
            test_origin.test.name = names_backup[i]

    def _remove_old_results(self, prefix: str) -> None:
        """
        Remove old results that were replaced by prefixed ones
        """

        results = self.step.plan.execute.results()
        old_result_names = [
            result.name.removeprefix(f'/{prefix}')
            for result in results
            if result.name.startswith(f'/{prefix}/')
        ]

        self.step.plan.execute._results = [
            result for result in results if result.name not in old_result_names
        ]
        self.step.plan.execute.save()



from collections.abc import Sequence
from typing import Optional

import tmt
import tmt.log
import tmt.steps
import tmt.steps.report
from tmt.container import container, field
from tmt.result import BaseResult, CheckResult, Result, SubCheckResult, SubResult
from tmt.steps.execute import TEST_OUTPUT_FILENAME
from tmt.utils import Path

# How much test and test check info should be shifted to the right in the output.
# We want tests to be shifted by one extra level, with their checks shifted by
# yet another level.
TEST_SHIFT = 1
CHECK_SHIFT = 2
SUBRESULT_SHIFT = CHECK_SHIFT
SUBRESULT_CHECK_SHIFT = SUBRESULT_SHIFT + 1


@container
class ReportDisplayData(tmt.steps.report.ReportStepData):
    display_guest: str = field(
        default='auto',
        option='--display-guest',
        metavar='auto|always|never',
        choices=['auto', 'always', 'never'],
        help="""
             When to display full guest name in report: when more than a single guest was involved
             (default), always, or never.
             """,
    )


@tmt.steps.provides_method('display')
class ReportDisplay(tmt.steps.report.ReportPlugin[ReportDisplayData]):
    """
    Show test results on the terminal.

    Give a concise summary of test results directly on the terminal.
    Allows to select the desired level of verbosity.

    .. code-block:: yaml

        tmt run -l report        # overall summary only
        tmt run -l report -v     # individual test results
        tmt run -l report -vv    # show full paths to logs
        tmt run -l report -vvv   # provide complete test output
    """

    _data_class = ReportDisplayData

    def details(self, result: tmt.Result, verbosity: int, display_guest: bool) -> None:
        """
        Print result details.

        :param result: a test result to display.
        :param verbosity: how verbose should the report be. Generally equal to
            number of  ``--verbose``/``-v`` options given on command line.
            For ``1``, display only test name and its result, ``2`` will add
            log paths, and ``3`` or more would show the test output as well.
        :param display_guest: if set, guest multihost name would be part of the
            report.
        """

        def _display_log_info(log: Path, shift: int) -> None:
            """
            Display info about a single result log
            """

            # TODO: are we sure it cannot be None?
            assert self.step.plan.execute.workdir is not None

            self.verbose(
                log.name, self.step.plan.execute.workdir / log, color='yellow', shift=shift
            )

        def _display_log_content(log: Path, shift: int) -> None:
            """
            Display content of a single result log
            """

            # TODO: are we sure it cannot be None?
            assert self.step.plan.execute.workdir is not None

            self.verbose(
                'content',
                self.read(self.step.plan.execute.workdir / log),
                color='yellow',
                shift=shift,
            )

        def display_outcome(result: BaseResult) -> None:
            """
            Display a single result outcome
            """

            if isinstance(result, SubCheckResult):
                self.verbose(
                    f'{result.show()} ({result.event.value} check)', shift=SUBRESULT_CHECK_SHIFT
                )

            elif isinstance(result, CheckResult):
                self.verbose(f'{result.show()} ({result.event.value} check)', shift=CHECK_SHIFT)

            elif isinstance(result, SubResult):
                self.verbose(result.show(), shift=SUBRESULT_SHIFT)

            elif isinstance(result, Result):
                self.verbose(result.show(display_guest=display_guest), shift=TEST_SHIFT)

        def display_log_info(result: BaseResult) -> None:
            """
            Display info about result logs
            """

            # TODO: are we sure it cannot be None?
            assert self.step.plan.execute.workdir is not None

            shift = (TEST_SHIFT + 1) if isinstance(result, Result) else (CHECK_SHIFT + 1)

            for log in result.log:
                _display_log_info(log, shift)

        def display_log_content(result: BaseResult) -> None:
            """
            Display content of interesting result logs
            """

            # TODO: are we sure it cannot be None?
            assert self.step.plan.execute.workdir is not None

            shift = (TEST_SHIFT + 1) if isinstance(result, Result) else (CHECK_SHIFT + 1)

            for log in result.log:
                _display_log_info(log, shift)

                if log.name == TEST_OUTPUT_FILENAME:
                    _display_log_content(log, shift)

        def display_subresults(results: Sequence[BaseResult]) -> None:
            """
            Display subresults, checks and subresult checks
            """

            for subresult in results:
                display_outcome(subresult)

                if verbosity > 2:
                    display_log_content(subresult)

                elif verbosity > 1:
                    display_log_info(subresult)

                # Recursively show also all the results of subresult checks
                if isinstance(subresult, SubResult):
                    display_subresults(subresult.check)

        # Always show the result outcome
        display_outcome(result)

        # With verbosity increased to `-vvv` or more, display content of the main test log
        if verbosity > 2:
            display_log_content(result)

        # With verbosity increased to `-vv`, display the list of logs
        elif verbosity > 1:
            display_log_info(result)

        display_subresults(result.check)
        display_subresults(result.subresult)

    def go(self, *, logger: Optional[tmt.log.Logger] = None) -> None:
        """
        Discover available tests
        """

        super().go(logger=logger)
        # Show individual test results only in verbose mode
        if not self.verbosity_level:
            return

        if self.data.display_guest == 'always':
            display_guest = True

        elif self.data.display_guest == 'never':
            display_guest = False

        else:
            seen_guests = {result.guest.name for result in self.step.plan.execute.results()}

            display_guest = len(seen_guests) > 1

        for result in self.step.plan.execute.results():
            self.details(result, self.verbosity_level, display_guest)



import os
import pickle
import shutil
import tempfile

import jsonschema
import pytest

import tmt
import tmt.cli._root
import tmt.result
from tests import CliRunner
from tmt.base import FmfId, Link, LinkNeedle, Links, expand_node_data
from tmt.utils import Path, SpecificationError

runner = CliRunner()


def test_invalid_yaml_syntax():
    """
    Invalid yaml syntax
    """

    tmp = tempfile.mkdtemp()
    original_directory = os.getcwd()
    os.chdir(tmp)
    result = runner.invoke(tmt.cli._root.main, ['init', '--template', 'mini'])
    with open('plans/example.fmf', 'a') as plan:
        plan.write('bad line')
    result = runner.invoke(tmt.cli._root.main)
    assert isinstance(result.exception, tmt.utils.GeneralError)
    assert result.exit_code != 0
    os.chdir(original_directory)
    shutil.rmtree(tmp)


def test_test_defaults(root_logger):
    """
    Test default test attributes
    """

    test = tmt.Test.from_dict(logger=root_logger, mapping={'test': './test.sh'}, name='/smoke')
    assert test.name == '/smoke'
    assert test.component == []
    assert str(test.test) == './test.sh'
    assert test.path == Path('/')
    assert test.require == []
    assert test.environment == {}
    assert test.duration == '5m'
    assert test.enabled is True
    assert test.result == tmt.result.ResultInterpret.RESPECT
    assert test.tag == []


def test_test_invalid(root_logger):
    """
    Test invalid test
    """

    # Missing name
    with pytest.raises(tmt.utils.GeneralError):
        tmt.Test.from_dict(logger=root_logger, mapping={}, name='')
    # Invalid name
    with pytest.raises(SpecificationError):
        tmt.Test.from_dict(logger=root_logger, mapping={}, name='bad')
    # Invalid attributes
    for key in ['component', 'require', 'tag']:
        with pytest.raises(SpecificationError) as exc_context:
            tmt.Test.from_dict(
                logger=root_logger, mapping={key: 1}, name='/smoke', raise_on_validation_error=True
            )

        exc = exc_context.value

        assert isinstance(exc, SpecificationError)
        assert exc.message == 'fmf node /smoke failed validation'

        validation_error, error_message = exc.validation_errors[0]

        assert isinstance(validation_error, jsonschema.ValidationError)
        assert error_message == f'/smoke:{key} - 1 is not valid under any of the given schemas'

    with pytest.raises(SpecificationError):
        tmt.Test.from_dict(
            logger=root_logger,
            mapping={'environment': 'string'},
            name='/smoke',
            raise_on_validation_error=True,
        )
    # Listify attributes
    assert tmt.Test.from_dict(
        logger=root_logger, mapping={'test': 'test', 'tag': 'a'}, name='/smoke'
    ).tag == ['a']
    assert tmt.Test.from_dict(
        logger=root_logger, mapping={'test': 'test', 'tag': ['a', 'b']}, name='/smoke'
    ).tag == ['a', 'b']


def test_link():
    """
    Test the link attribute parsing
    """

    # No link should default to an empty list
    assert Links().get() == []

    # Single string (default relation)
    assert Links(data='/fmf/id').get() == [Link(relation='relates', target='/fmf/id')]
    # Multiple strings (default relation)
    assert Links(data=['one', 'two']).get() == [
        Link(relation='relates', target='one'),
        Link(relation='relates', target='two'),
    ]
    # Multiple string mixed relation
    assert Links(data=['implicit', {'duplicates': 'explicit'}]).get() == [
        Link(relation='relates', target='implicit'),
        Link(relation='duplicates', target='explicit'),
    ]
    # Multiple strings (explicit relation)
    assert Links(data=[{'parent': 'mom'}, {'child': 'son'}]).get() == [
        Link(relation='parent', target='mom'),
        Link(relation='child', target='son'),
    ]

    # Single dictionary (default relation)
    assert Links(data={'name': 'foo'}).get() == [
        Link(relation='relates', target=FmfId(name='foo'))
    ]
    # Single dictionary (explicit relation)
    assert Links(data={'verifies': 'foo'}).get() == [Link(relation='verifies', target='foo')]
    # Multiple dictionaries
    family = [{'parent': 'mom', 'note': 'foo'}, {'child': 'son'}]
    assert Links(data=family).get() == [
        Link(relation='parent', target='mom', note='foo'),
        Link(relation='child', target='son'),
    ]

    # Selected relations
    assert Links(data=family).get('parent') == [Link(relation='parent', target='mom', note='foo')]
    assert Links(data=family).get('child') == [Link(relation='child', target='son')]

    # Full fmf id
    fmf_id = tmt.utils.yaml_to_dict("""
        blocked-by:
            url: https://github.com/teemtee/fmf
            name: /stories/select/filter/regexp
        note: Need to get the regexp filter working first.
        """)
    link = Links(data=fmf_id)
    assert link.get() == [
        Link(
            relation='blocked-by',
            target=FmfId(url=fmf_id['blocked-by']['url'], name=fmf_id['blocked-by']['name']),
            note=fmf_id['note'],
        )
    ]

    # Invalid links and relations
    with pytest.raises(
        SpecificationError,
        match="Field 'link' must be a string, a fmf id or a list of their combinations,"
        " 'int' found.",
    ):
        Links(data=123)
    with pytest.raises(SpecificationError, match='Multiple relations'):
        Links(data={'verifies': 'one', 'blocks': 'another'})
    with pytest.raises(SpecificationError, match='Invalid link relation'):
        Links(data={'depends': 'other'})

    # Searching for links
    links = Links(data=[{'parent': 'mom', 'note': 'foo'}, {'child': 'son', 'note': 'bar'}])
    assert links.has_link()
    assert links.has_link(needle=LinkNeedle())
    assert links.has_link(needle=LinkNeedle(relation=r'.*', target=r'.*'))
    assert links.has_link(needle=LinkNeedle(relation='[a-z]+'))
    assert links.has_link(needle=LinkNeedle(relation='en'))
    assert links.has_link(needle=LinkNeedle(target='^mom$'))
    assert links.has_link(needle=LinkNeedle(target='on'))
    assert not links.has_link(needle=LinkNeedle(relation='verifies', target='son'))
    assert not links.has_link(needle=LinkNeedle(relation='parent', target='son'))

    links = Links(data=[])
    assert not links.has_link()
    assert not links.has_link(needle=LinkNeedle())
    assert not links.has_link(needle=LinkNeedle(relation=r'.*', target=r'.*'))


def test_pickleable_tree() -> None:
    """
    https://github.com/teemtee/tmt/issues/2503
    """

    tree = tmt.Tree.grow()

    pickle.dumps(tree)


def test_expand_node_data(monkeypatch) -> None:
    """
    :py:func:`tmt.base.expand_node_data` handles various forms of variables
    """

    # From ``_data` and `_expected` we construct lists with items, including
    # another list and a dictionary, to verify `expand_node_data()` handles
    # nested structures too.
    #
    # Then we will call `expand_node_data()` with custom fmf context and
    # environment providing inputs.

    fmf_context = {'WALDO': ['plugh'], 'XYZZY': ['thud']}

    environ = {
        'CORGE': 'quuz',
        # envvars starting with `@` resulted in `$@...` being the outcome,
        # as described in https://github.com/teemtee/tmt/issues/2654
        'FRED': '@XYZZY',
    }

    # All values should be propagated as they are, unless comment describes the expected
    # transformation.
    _data = [
        'foo',
        1,
        True,
        # `$CORGE` should be replaced with `quuz` from the environment
        '$CORGE',
        '  $CORGE  ',
        'something${CORGE}else',
        # `$FRED` should be replaced with `@XYZZY` from the environment
        # Similar to `$CORGE`, but testing whether leading `@` is handled
        # properly, see https://github.com/teemtee/tmt/issues/2654
        '$FRED',
        '  $FRED  ',
        'something${FRED}else',
        # `@WALDO` should be replaced by `plugh` from fmf context
        '$@WALDO',
        '  $@WALDO  ',
        'something$@{WALDO}else',
        # There is no source for these, therefore they remain untouched
        '$GRAPLY',
        '  $GRAPLY  ',
        'something${GRAPLY}else',
        '$@GRAPLY',
        '  $@GRAPLY  ',
        'something$@{GRAPLY}else',
    ]

    _expected = [
        'foo',
        1,
        True,
        'quuz',
        '  quuz  ',
        'somethingquuzelse',
        '@XYZZY',
        '  @XYZZY  ',
        'something@XYZZYelse',
        'plugh',
        '  plugh  ',
        'somethingplughelse',
        '$GRAPLY',
        '  $GRAPLY  ',
        'something${GRAPLY}else',
        '$@GRAPLY',
        '  $@GRAPLY  ',
        'something$@{GRAPLY}else',
    ]

    data = [*_data, [*_data], {f'key{i}': value for i, value in enumerate(_data)}]

    expected = [*_expected, [*_expected], {f'key{i}': value for i, value in enumerate(_expected)}]

    for envvar in os.environ:
        monkeypatch.delenv(envvar)

    for envvar, value in environ.items():
        monkeypatch.setenv(envvar, value)

    assert expand_node_data(data, fmf_context) == expected



"""
Common options and the MethodCommand class
"""

import contextlib
import pathlib
import re
import textwrap
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, Callable, Optional, Union

import click

import tmt.lint
import tmt.log
import tmt.utils
from tmt.container import container

# When dealing with older Click packages (I'm looking at you, Python 3.6),
# we need to define FC on our own.
try:
    from click.decorators import FC

except ImportError:
    from typing import TypeVar

    FC = TypeVar('FC', bound=Union[Callable[..., Any], click.Command])  # type: ignore[misc]


if TYPE_CHECKING:
    import tmt.cli
    import tmt.utils


@container(frozen=True)
class Deprecated:
    """
    Version information and hint for obsolete options
    """

    since: str
    hint: Optional[str] = None

    @property
    def rendered(self) -> str:
        message = f"The option is deprecated since {self.since}"

        if self.hint:
            message = f'{message}, {self.hint}'

        return f'{message}.'


class Path(click.ParamType):
    name = 'path'

    def convert(
        self,
        value: Any,
        param: Optional[click.Parameter],
        ctx: Optional[click.Context],
    ) -> Optional[tmt.utils.Path]:
        """
        Convert the value to the correct type. This is not called if
        the value is ``None`` (the missing value).

        This must accept string values from the command line, as well as
        values that are already the correct type. It may also convert
        other compatible types.

        The ``param`` and ``ctx`` arguments may be ``None`` in certain
        situations, such as when converting prompt input.

        If the value cannot be converted, call :meth:`fail` with a
        descriptive message.

        :param value: The value to convert.
        :param param: The parameter that is using this type to convert
            its value. May be ``None``.
        :param ctx: The current context that arrived at this value. May
            be ``None``.
        """

        if value is None:
            return None

        if isinstance(value, str):
            return tmt.utils.Path(value)

        if isinstance(value, tmt.utils.Path):
            return value

        if isinstance(value, pathlib.Path):  # noqa: TID251
            return tmt.utils.Path(str(value))

        if param is None:
            self.fail(
                f"A path-like string was expected, '{type(value).__name__}' found.",
                param=param,
                ctx=ctx,
            )

        # RET503: ruff does not recognize NoReturn annotation of `self.fail`.
        self.fail(  # noqa: RET503
            f"Field '{param.name}' must be a path-like string, '{type(value).__name__}' found.",
            param=param,
            ctx=ctx,
        )


MethodDictType = dict[str, click.core.Command]

# Originating in click.decorators, an opaque type describing "decorator" functions
# produced by click.option() calls: not options, but decorators, functions that attach
# options to a given command.
# Since click.decorators does not have a dedicated type for this purpose, we need
# to construct it on our own, but we can reuse a typevar click.decorators has.
_ClickOptionDecoratorType = Callable[[FC], FC]
# The type above is a generic type, `FC` being a typevar, so we have two options:
# * each place using the type would need to fill the variable, i.e. add [foo]`, or
# * we could do that right here, because right now, we don't care too much about
# what this `foo` type actually is - what's important is the identity, return type
# matches the type of the argument.
ClickOptionDecoratorType = _ClickOptionDecoratorType[Any]


def option(
    *param_decls: str,
    # Following parameters are inherited from click.option()/Option/Parameter.
    # May allow stricter types than the original, because tmt code base does not
    # care for every Click use case.
    show_default: bool = False,
    is_flag: bool = False,
    multiple: bool = False,
    count: bool = False,
    type: Optional[Union[click.Choice, Any]] = None,  # noqa: A002  `type` is shadowing a Python builtin
    help: Optional[str] = None,
    required: bool = False,
    default: Optional[Any] = None,
    nargs: Optional[int] = None,
    metavar: Optional[str] = None,
    prompt: Optional[str] = None,
    envvar: Optional[str] = None,
    hidden: bool = False,
    # Following parameters are our additions.
    choices: Optional[Sequence[str]] = None,
    deprecated: Optional[Deprecated] = None,
) -> ClickOptionDecoratorType:
    """
    Attaches an option to the command.

    This is a wrapper for :py:func:`click.option`, its parameters have the same
    meaning as those of ``click.option()``, and are passed to ``click.option()``,
    with the exception of ``deprecated`` parameter.

    :param choices: if set, it sets ``type`` of the option to
        :py:class:`click.Choices`, and limits option values to those
        listed in ``choices``.
    :param deprecated: if set, it is rendered and appended to ``help``. This
        parameter is **not** passed to :py:func:`click.option`.
    """

    if help:
        help = textwrap.dedent(help)

    # Add a deprecation warning for obsoleted options
    if deprecated:
        if help:
            help += ' ' + deprecated.rendered

        else:
            help = deprecated.rendered

    if choices is not None:
        type = click.Choice(choices)  # noqa: A001  `type` is shadowing a Python builtin

    # Add a metavar listing choices unless an explicit metavar has been provided
    if isinstance(type, click.Choice) and metavar is None:
        metavar = '|'.join(type.choices)

    # Instead of repeating all keyword parameters, use locals(), they are all there
    # already, and it's a dictionary - just don't forget to remove names that are
    # not accepted by click and the positional parameter.
    kwargs = locals().copy()
    kwargs.pop('param_decls')
    kwargs.pop('choices')
    kwargs.pop('deprecated')

    return click.option(*param_decls, **kwargs)


# Verbose, debug and quiet output
VERBOSITY_OPTIONS: list[ClickOptionDecoratorType] = [
    option(
        '-v',
        '--verbose',
        count=True,
        default=0,
        help='Show more details. Use multiple times to raise verbosity.',
    ),
    option(
        '-d',
        '--debug',
        count=True,
        default=0,
        help='Provide debugging information. Repeat to see more details.',
    ),
    option(
        '-q',
        '--quiet',
        is_flag=True,
        help='Be quiet. Exit code is just enough for me.',
    ),
    option(
        '--log-topic',
        choices=[topic.value for topic in tmt.log.Topic],
        multiple=True,
        help='If specified, --debug and --verbose would emit logs also for these topics.',
    ),
]

# Force, dry, feeling-safe and run again actions
DRY_OPTIONS: list[ClickOptionDecoratorType] = [
    option(
        '-n',
        '--dry',
        is_flag=True,
        default=False,
        help='Run in dry mode. No changes, please.',
    ),
]

FORCE_DRY_OPTIONS: list[ClickOptionDecoratorType] = [
    option(
        '-f',
        '--force',
        is_flag=True,
        help='Overwrite existing files and step data.',
    ),
    *DRY_OPTIONS,
]

AGAIN_OPTION: list[ClickOptionDecoratorType] = [
    option(
        '--again',
        is_flag=True,
        help='Run again, even if already done before.',
    ),
]

FEELING_SAFE_OPTION: list[ClickOptionDecoratorType] = [
    option(
        '--feeling-safe',
        metavar='FEELING_SAFE',
        envvar='TMT_FEELING_SAFE',
        is_flag=True,
        default=False,
        help="""
             WARNING: with this option, tmt would be allowed to make
             potentially dangerous actions. For example, some metadata
             keys may cause scripts being executed on the runner.
             Do not use this option unless you trust metadata consumed
             by tmt, or unless you know what you are doing.
             """,
    )
]

# Fix action
FIX_OPTIONS: list[ClickOptionDecoratorType] = [
    option('-F', '--fix', is_flag=True, help='Attempt to fix all discovered issues.')
]

WORKDIR_ROOT_OPTIONS: list[ClickOptionDecoratorType] = [
    option(
        '--workdir-root',
        metavar='PATH',
        envvar='TMT_WORKDIR_ROOT',
        default=tmt.utils.WORKDIR_ROOT,
        type=Path(),
        help=f"""
             Path to root directory containing run workdirs.
             Defaults to '{tmt.utils.WORKDIR_ROOT}'.
             """,
    )
]

FILTER_OPTION: list[ClickOptionDecoratorType] = [
    option(
        '-f',
        '--filter',
        'filters',
        metavar='FILTER',
        multiple=True,
        help="""
        Apply an advanced filter using key:value pairs and logical operators.
        For example 'tier:1 & tag:core'. Use the 'name' key to search by name.
        See 'pydoc fmf.filter' for detailed documentation on the syntax.
        """,
    ),
]

FILTER_OPTION_LONG: list[ClickOptionDecoratorType] = [
    option(
        '--filter',
        'filters',
        metavar='FILTER',
        multiple=True,
        help="""
        Apply an advanced filter using key:value pairs and logical operators.
        For example 'tier:1 & tag:core'. Use the 'name' key to search by name.
        See 'pydoc fmf.filter' for detailed documentation on the syntax.
        """,
    ),
]

FILTERING_OPTIONS: list[ClickOptionDecoratorType] = [
    click.argument('names', nargs=-1, metavar='[REGEXP|.]'),
    *FILTER_OPTION,
    option(
        '-c',
        '--condition',
        'conditions',
        metavar="EXPR",
        multiple=True,
        help="Use arbitrary Python expression for filtering.",
    ),
    option(
        '--enabled',
        is_flag=True,
        help="Show only enabled tests, plans or stories.",
    ),
    option(
        '--disabled',
        is_flag=True,
        help="Show only disabled tests, plans or stories.",
    ),
    option(
        '--link',
        'links',
        metavar="RELATION:TARGET",
        multiple=True,
        help="""
             Filter by linked objects (regular expressions are supported for both relation and
             target).
             """,
    ),
    option(
        '-x',
        '--exclude',
        'exclude',
        metavar='[REGEXP]',
        multiple=True,
        help="Exclude a regular expression from search result.",
    ),
]


FILTERING_OPTIONS_LONG: list[ClickOptionDecoratorType] = [
    click.argument('names', nargs=-1, metavar='[REGEXP|.]'),
    *FILTER_OPTION_LONG,
    option(
        '--condition',
        'conditions',
        metavar="EXPR",
        multiple=True,
        help="Use arbitrary Python expression for filtering.",
    ),
    option(
        '--enabled',
        is_flag=True,
        help="Show only enabled tests, plans or stories.",
    ),
    option(
        '--disabled',
        is_flag=True,
        help="Show only disabled tests, plans or stories.",
    ),
    option(
        '--link',
        'links',
        metavar="RELATION:TARGET",
        multiple=True,
        help="""
             Filter by linked objects (regular expressions are supported for both relation and
             target).
             """,
    ),
    option(
        '--exclude',
        'exclude',
        metavar='[REGEXP]',
        multiple=True,
        help="Exclude a regular expression from search result.",
    ),
]


STORY_FLAGS_FILTER_OPTIONS: list[ClickOptionDecoratorType] = [
    option(
        '--implemented',
        is_flag=True,
        help='Implemented stories only.',
    ),
    option(
        '--unimplemented',
        is_flag=True,
        help='Unimplemented stories only.',
    ),
    option(
        '--verified',
        is_flag=True,
        help='Stories verified by tests.',
    ),
    option(
        '--unverified',
        is_flag=True,
        help='Stories not verified by tests.',
    ),
    option(
        '--documented',
        is_flag=True,
        help='Documented stories only.',
    ),
    option(
        '--undocumented',
        is_flag=True,
        help='Undocumented stories only.',
    ),
    option(
        '--covered',
        is_flag=True,
        help='Covered stories only.',
    ),
    option(
        '--uncovered',
        is_flag=True,
        help='Uncovered stories only.',
    ),
]

FMF_SOURCE_OPTIONS: list[ClickOptionDecoratorType] = [
    option(
        '--source',
        is_flag=True,
        help="Select by fmf source file names instead of object names.",
    )
]

REMOTE_PLAN_OPTIONS: list[ClickOptionDecoratorType] = [
    option(
        '-s',
        '--shallow',
        is_flag=True,
        help='Do not clone remote plan.',
    )
]


_lint_outcomes = [member.value for member in tmt.lint.LinterOutcome.__members__.values()]

LINT_OPTIONS: list[ClickOptionDecoratorType] = [
    option(
        '--list-checks',
        is_flag=True,
        help='List all available checks.',
    ),
    option(
        '--enable-check',
        'enable_checks',
        metavar='CHECK-ID',
        multiple=True,
        type=str,
        help='Run only checks mentioned by this option.',
    ),
    option(
        '--disable-check',
        'disable_checks',
        metavar='CHECK-ID',
        multiple=True,
        type=str,
        help='Do not run checks mentioned by this option.',
    ),
    option(
        '--enforce-check',
        'enforce_checks',
        metavar='CHECK-ID',
        multiple=True,
        type=str,
        help='Consider linting as failed if any of the mentioned checks is not a pass.',
    ),
    option(
        '--failed-only',
        is_flag=True,
        help='Display only tests/plans/stories that fail a check.',
    ),
    option(
        '--outcome-only',
        multiple=True,
        choices=_lint_outcomes,
        help='Display only checks with the given outcome.',
    ),
]


ENVIRONMENT_OPTIONS: list[ClickOptionDecoratorType] = [
    option(
        '-e',
        '--environment',
        metavar='KEY=VALUE|@FILE',
        multiple=True,
        help="""
             Set environment variable. Can be specified multiple times. The "@" prefix marks a file
             to load (yaml or dotenv formats supported).
             """,
    ),
    option(
        '--environment-file',
        metavar='FILE|URL',
        multiple=True,
        help="""
             Set environment variables from file or url (yaml or dotenv formats are supported). Can
             be specified multiple times.
             """,
    ),
]


def create_options_decorator(options: list[ClickOptionDecoratorType]) -> Callable[[FC], FC]:
    def common_decorator(fn: FC) -> FC:
        for option in reversed(options):
            fn = option(fn)

        return fn

    return common_decorator


def create_method_class(methods: MethodDictType) -> type[click.Command]:
    """
    Create special class to handle different options for each method

    Accepts dictionary with method names and corresponding commands:
    For example: {'fmf', <click.core.Command object at 0x7f3fe04fded0>}
    Methods should be already sorted according to their priority.
    """

    def is_likely_subcommand(arg: str, subcommands: list[str]) -> bool:
        """
        Return true if arg is the beginning characters of a subcommand
        """

        return any(subcommand.startswith(arg) for subcommand in subcommands)

    class MethodCommand(click.Command):
        _method: Optional[click.Command] = None

        def _check_method(self, context: 'tmt.cli.Context', args: list[str]) -> None:
            """
            Manually parse the --how option
            """

            # Avoiding circular imports
            import tmt.steps

            # TODO: this one is weird: `tmt.utils` is already imported on module
            # level, yet pyright believes `"utils" is not a known member of module
            # "tmt"`. Maybe there's some circular import, but I've been unable to
            # find it.
            import tmt.utils

            how = None
            subcommands = tmt.steps.STEPS + tmt.steps.ACTIONS + ['tests', 'plans']

            def _find_option_by_arg(arg: str) -> Optional[click.Parameter]:
                for option in self.params:
                    if arg in option.opts or arg in option.secondary_opts:
                        return option
                return None

            def _find_how(args: list[str]) -> Optional[str]:
                while args:
                    arg = args.pop(0)

                    # Handle '--how method' or '-h method'
                    if arg in ['--how', '-h']:
                        # Found `-h/--how foo`, next argument is how'
                        return args.pop(0)

                    # Handle '--how=method'
                    if arg.startswith('--how='):
                        return re.sub('^--how=', '', arg)

                    # Handle '-hmethod'
                    if arg.startswith('-h'):
                        return re.sub('^-h ?', '', arg)

                    # Handle anything that looks like an option
                    if arg.startswith('-'):
                        # Is it a --foo or --foo=bar?
                        match = re.match(r'(--[a-z\-]+)(=.*)?', arg)
                        if match:
                            if match.group(2):
                                # Found option like '--foo=bar'
                                continue
                        else:
                            # Or is it a -f?
                            match = re.match(r'(-[a-z])', arg)
                            if match is None:
                                # Found unexpected option format
                                return None
                        option_name: str = match.group(1)
                        option = _find_option_by_arg(option_name)
                        if option is None:
                            # Unknown option? Probably remain silent, Click should report it in.
                            return None

                        # Options, unlike arguments, may not consume any additional arguments
                        if isinstance(option, click.core.Option):
                            if option.is_flag:
                                # Found a flag
                                continue
                            if option.count:
                                # Found options like '-ddddd'
                                continue

                        # Consume all remaining arguments. Think `ls /foo /bar ...`, it's not
                        # possible to tell which of these is an actual argument or misspelled
                        # subcommand name.
                        if option.nargs == -1:
                            # Found option consuming all arguments
                            return None
                        if option.nargs == 0:
                            # Found option with no arguments
                            continue

                        # Consume the given number of arguments
                        for _ in range(option.nargs):
                            args.pop(0)

                        continue

                    if is_likely_subcommand(arg, subcommands):
                        # Found a subcommand
                        return None

                    # We're left with a string that does not start with `-`, and which has not
                    # been claimed by an option. It is highly likely this is a misspelled
                    # subcommand name.
                    raise tmt.utils.SpecificationError(
                        f"Invalid subcommand of 'run' found: '{arg}'."
                    )

                return None

            with contextlib.suppress(IndexError):
                how = _find_how(args[:])

            # Find method with the first matching prefix
            if how is not None:
                for method in methods:
                    if method.startswith(how):
                        self._method = methods[method]
                        break

            if how and self._method is None:
                from tmt.utils.hints import print_hints

                # Use run for logging, steps may not be initialized yet
                assert context.obj.run is not None  # narrow type
                assert self.name is not None  # narrow type

                print_hints(
                    f'{self.name}/{how}', ignore_missing=True, logger=context.obj.run._logger
                )
                print_hints(self.name, ignore_missing=True, logger=context.obj.run._logger)

                raise tmt.utils.SpecificationError(f"Unsupported {self.name} method '{how}'.")

        def parse_args(  # type: ignore[override]
            self, context: 'tmt.cli.Context', args: list[str]
        ) -> list[str]:
            self._check_method(context, args)
            if self._method is not None:
                return self._method.parse_args(context, args)
            return super().parse_args(context, args)

        def get_help(self, context: 'tmt.cli.Context') -> str:  # type: ignore[override]
            if self._method is not None:
                return self._method.get_help(context)
            return super().get_help(context)

        def invoke(self, context: 'tmt.cli.Context') -> Any:  # type: ignore[override]
            if self._method:
                return self._method.invoke(context)
            return super().invoke(context)

    return MethodCommand



import os

import pytest

env_vars_parametrization = (
    ("env_name", "value"),
    (
        ("STR", "O"),
        ("INT", "0"),
        ("DOTENV_STR", "dotenv_str"),
        ("DOTENV_INT", "1"),
        ("YAML_STR", "yaml_str_overwritten"),
        ("YAML_INT", "1"),
        ("SPECIAL", "/=(;-)"),
    ),
)


@pytest.mark.parametrize(*env_vars_parametrization)
def test_environment_file_with_variables(env_name, value):
    assert os.environ[env_name] == value



from typing import Any, Optional, cast

import fmf
import fmf.utils

import tmt
import tmt.log
import tmt.steps
import tmt.steps.prepare
import tmt.utils
from tmt.container import container, field
from tmt.result import PhaseResult
from tmt.steps import safe_filename
from tmt.steps.provision import Guest
from tmt.utils import ShellScript

PREPARE_WRAPPER_FILENAME = 'tmt-prepare-wrapper.sh'


@container
class PrepareShellData(tmt.steps.prepare.PrepareStepData):
    script: list[ShellScript] = field(
        default_factory=list,
        option=('-s', '--script'),
        multiple=True,
        metavar='SCRIPT',
        help='Shell script to be executed. Can be used multiple times.',
        normalize=tmt.utils.normalize_shell_script_list,
        serialize=lambda scripts: [str(script) for script in scripts],
        unserialize=lambda serialized: [ShellScript(script) for script in serialized],
    )

    # ignore[override] & cast: two base classes define to_spec(), with conflicting
    # formal types.
    def to_spec(self) -> dict[str, Any]:  # type: ignore[override]
        data = cast(dict[str, Any], super().to_spec())
        data['script'] = [str(script) for script in self.script]

        return data


@tmt.steps.provides_method('shell')
class PrepareShell(tmt.steps.prepare.PreparePlugin[PrepareShellData]):
    """
    Prepare guest using shell (Bash) scripts.

    Default shell options are applied to the script, see the
    :ref:`/spec/tests/test` key specification for more
    details.

    .. code-block:: yaml

        prepare:
            how: shell
            script:
              - sudo dnf install -y 'dnf-command(copr)'
              - sudo dnf copr enable -y psss/tmt
              - sudo dnf install -y tmt
    """

    _data_class = PrepareShellData

    def go(
        self,
        *,
        guest: 'Guest',
        environment: Optional[tmt.utils.Environment] = None,
        logger: tmt.log.Logger,
    ) -> list[PhaseResult]:
        """
        Prepare the guests
        """

        results = super().go(guest=guest, environment=environment, logger=logger)

        environment = environment or tmt.utils.Environment()

        # Give a short summary
        overview = fmf.utils.listed(self.data.script, 'script')
        logger.info('overview', f'{overview} found', 'green')

        workdir = self.step.plan.worktree
        assert workdir is not None  # narrow type

        if not self.is_dry_run:
            topology = tmt.steps.Topology(self.step.plan.provision.ready_guests)
            topology.guest = tmt.steps.GuestTopology(guest)

            environment.update(
                topology.push(
                    dirpath=workdir,
                    guest=guest,
                    logger=logger,
                    filename_base=safe_filename(
                        tmt.steps.TEST_TOPOLOGY_FILENAME_BASE, self, guest
                    ),
                )
            )

        prepare_wrapper_filename = safe_filename(PREPARE_WRAPPER_FILENAME, self, guest)
        prepare_wrapper_path = workdir / prepare_wrapper_filename

        logger.debug('prepare wrapper', prepare_wrapper_path, level=3)

        # Execute each script on the guest (with default shell options)
        for script in self.data.script:
            logger.verbose('script', script, 'green')
            script_with_options = tmt.utils.ShellScript(f'{tmt.utils.SHELL_OPTIONS}; {script}')
            self.write(prepare_wrapper_path, str(script_with_options), 'w')
            if not self.is_dry_run:
                prepare_wrapper_path.chmod(0o755)
            guest.push(
                source=prepare_wrapper_path,
                destination=prepare_wrapper_path,
                options=["-s", "-p", "--chmod=755"],
            )
            command: ShellScript
            if guest.become and not guest.facts.is_superuser:
                command = tmt.utils.ShellScript(f'sudo -E {prepare_wrapper_path}')
            else:
                command = tmt.utils.ShellScript(f'{prepare_wrapper_path}')
            guest.execute(command=command, cwd=workdir, env=environment)

        return results



"""
``tmt init`` implementation
"""

from typing import Any

import click

import tmt.templates
from tmt.cli import Context, pass_context
from tmt.cli._root import force_dry_options, main, verbosity_options
from tmt.options import option
from tmt.utils import Path


@main.command()
@pass_context
@click.argument('path', default='.')
@option(
    '-t',
    '--template',
    default='empty',
    choices=['empty', *tmt.templates.INIT_TEMPLATES],
    help="Use this template to populate the tree.",
)
@verbosity_options
@force_dry_options
def init(
    context: Context,
    path: str,
    template: str,
    force: bool,
    **kwargs: Any,
) -> None:
    """
    Initialize a new tmt tree.

    By default tree is created in the current directory.
    Provide a PATH to create it in a different location.

    \b
    A tree can be optionally populated with example metadata:
    * 'mini' template contains a minimal plan and no tests,
    * 'base' template contains a plan and a beakerlib test,
    * 'full' template contains a 'full' story, an 'full' plan and a shell test.
    """

    tmt.Tree.store_cli_invocation(context)
    tmt.Tree.init(logger=context.obj.logger, path=Path(path), template=template, force=force)



import tmt
import tmt.steps
from tmt.options import option

# See the online documentation for more details about writing plugins
# https://tmt.readthedocs.io/en/stable/plugins.html


@tmt.steps.provides_method('example')
class ProvisionExample(tmt.steps.provision.ProvisionPlugin):
    """
    Provision guest using nothing. Just example

    Minimal configuration using the latest nothing image:

    .. code-block:: yaml

        provision:
            how: example

    Full configuration example:

    .. code-block:: yaml

        provision:
            how: example
    """

    _guest = None

    @classmethod
    def options(cls, how=None):
        """
        Prepare command line options for example
        """

        return [
            option(
                '-w',
                '--what',
                metavar='WHAT',
                help="Example how to pass value.",
            ),
            option(
                '-s',
                '--switch',
                is_flag=True,
                help="Example how to enable something.",
            ),
            *super().options(how),
        ]

    def default(self, option, default=None):
        """
        Return the default value for the given option
        """

        defaults = {
            'what': 'default value',
            'switch': False,
        }
        return defaults.get(option, default)

    def show(self):
        """
        Show provision details
        """

        super().show(['what', 'switch'])

    def wake(self, data=None):
        """
        Wake up the plugin

        Override data with command line options.
        Wake up the guest based on provided guest data.
        """

        super().wake(['what', 'switch'])
        print("wake() called")

        # Don't schedule anything if ve are in dry mode
        if self.opt('dry'):
            return

        if data:
            self._guest = GuestExample(data, name=self.name, parent=self.step)
            self._guest.wake()

    def go(self):
        """
        Provision the container
        """

        super().go()
        print("go() called")

        # Data dictionary is used to pass information among classes.
        data = {'what': 'Another default what. Object variable can be used.'}

        for opt in ['what', 'switch']:
            val = self.get(opt)
            # You can hide some not important information about provisioning.
            if opt != 'switch':
                self.info(opt, val, 'green')
            data[opt] = val

        self._guest = GuestExample(data, name=self.name, parent=self.step)
        self._guest.start()
        self._guest.setup()

    def guest(self):
        """
        Return provisioned guest

        Each ProvisionPlugin has to implement this method.
        Should return a provisioned Guest() instance.
        """

        return self._guest


class GuestExample(tmt.Guest):
    """
    Guest provisioned for test execution

    The following keys are expected in the 'data' dictionary::

    guest ...... hostname or ip address
    port ....... port to connect to
    user ....... user name to log in
    key ........ private key
    password ... password
    become ..... whether to run the scripts with sudo

    These are by default imported into instance attributes (see the
    class attribute '_keys' in tmt.Guest class).

    Example instance:

    The following keys are expected in the 'data' dictionary:

        what ...... content of cli what
        switch .... flag passed via cli
    """

    def load(self, data):
        """
        Load guest data into object attributes for easy access

        Called during guest object initialization. Takes care of storing
        all supported keys (see class attribute _keys in tmt.Guest class
        for the list) from provided data to the guest object attributes.
        Child classes can extend it to make additional guest attributes
        easily available.

        Data dictionary can contain guest information from both command
        line options / L2 metadata / user configuration and wake up data
        stored by the save() method below.
        """

        super().load(data)
        self.what = data.get('what')
        self.switch = data.get('switch')

    def wake(self):
        """
        Wake up the guest

        Perform any actions necessary after step wake up to be able to
        attach to a running guest instance and execute commands. Called
        after load() is completed so all guest data should be prepared.
        """

        print("wake() called")

    def save(self):
        """
        Save guest data for future wake up

        Export all essential guest data into a dictionary which will be
        stored in the `guests.yaml` file for possible future wake up of
        the guest. Everything needed to attach to a running instance
        should be added into the data dictionary by child classes.
        """

        data = super().save()
        data['what'] = self.what
        data['switch'] = self.switch
        return data

    def _some_your_internal_stuff(self):
        """
        Do some heavy lifting
        """

        return True

    def start(self):
        """
        Start the guest

        Get a new guest instance running. This should include preparing
        any configuration necessary to get it started. Called after
        load() is completed so all guest data should be available.
        """

        print("start() called")

        if self.opt('dry'):
            return

        self.verbose('what', self.what, 'green')

        if self._some_your_internal_stuff():
            return

        raise tmt.utils.ProvisionError("All attempts to provision a machine with example failed.")

    def setup(self):
        """
        Setup the guest

        This should include all necessary configurations inside the instance
        to get the plans to work. For example, ensure the permissions in
        TMT_WORKDIR_ROOT will work if the user is non-root.
        """

        print("setup() called")

    # For advanced development
    def execute(self, command, **kwargs):
        """
        Optionally you can overload how commands going to be executed
        on guest (provisioned machine). If you don't want to use
        ssh to connect to guest, you need to overload this method
        however you need to provide some expected information.

        Execute command on the guest

            command ... string or list of command arguments (required)
            env ....... dictionary with environment variables
            cwd ....... working directory to be entered before execution

        If the command is provided as a list, it will be space-joined.
        If necessary, quote escaping has to be handled by the caller.
        """

        print("execute() called. This is an optional overload...")

        return ["Fedora", "whatever"]

    def delete(self):
        """
        Remove the example instance
        """

        self.debug("You should place code for cleanup here.")

    def remove(self):
        """
        Remove the guest
        """

        if self.what:
            self.info('guest', 'removed', 'green')
            self.delete()
            self.what = None



import tmt
import tmt.log
import tmt.plugins
import tmt.steps.discover
import tmt.templates
import tmt.utils
from tmt.utils import Path


def test_init_custom_templates_folder(tmppath: Path) -> None:
    templates_dir = tmppath / 'templates'
    template_manager = tmt.templates.TemplateManager(templates_dir)

    story_dir = templates_dir / 'story'
    plan_dir = templates_dir / 'plan'
    test_metadata_dir = templates_dir / 'test'
    test_script_dir = templates_dir / 'script'

    assert template_manager.custom_template_path == templates_dir
    assert templates_dir.exists()
    assert story_dir.exists()
    assert plan_dir.exists()
    assert test_script_dir.exists()
    assert test_metadata_dir.exists()


def test_get_default_templates(tmppath: Path) -> None:
    template_manager = tmt.templates.TemplateManager(tmppath)

    default_templates = template_manager.default_templates['default']
    story_templates = template_manager.default_templates['story']
    plan_templates = template_manager.default_templates['plan']
    test_metadata_templates = template_manager.default_templates['test']
    test_script_templates = template_manager.default_templates['script']

    assert default_templates['plan'] is not None
    assert story_templates['full'] is not None
    assert plan_templates['full'] is not None
    assert test_script_templates['shell'] is not None
    assert test_metadata_templates['shell'] is not None


def test_get_custom_templates(tmppath: Path) -> None:
    def create_template_file(template_path: Path, content: str) -> None:
        template_path.write_text(content)

    template_manager = tmt.templates.TemplateManager(tmppath)

    create_template_file(
        template_manager.custom_template_path / 'story' / 'tmp_story.j2', 'Story template content.'
    )
    create_template_file(
        template_manager.custom_template_path / 'plan' / 'tmp_plan.j2', 'Plan template content.'
    )
    create_template_file(
        template_manager.custom_template_path / 'test' / 'tmp_test_metadata.j2',
        'Test metadata template content.',
    )
    create_template_file(
        template_manager.custom_template_path / 'script' / 'tmp_test_script.j2',
        'Test script template content.',
    )

    assert (
        template_manager.custom_templates['story']['tmp_story'].read_text()
        == 'Story template content.'
    )
    assert (
        template_manager.custom_templates['plan']['tmp_plan'].read_text()
        == 'Plan template content.'
    )
    assert (
        template_manager.custom_templates['test']['tmp_test_metadata'].read_text()
        == 'Test metadata template content.'
    )
    assert (
        template_manager.custom_templates['script']['tmp_test_script'].read_text()
        == 'Test script template content.'
    )


def test_get_combined_templates(tmppath: Path) -> None:
    def create_template_file(template_path: Path, content: str) -> None:
        template_path.write_text(content)

    template_manager = tmt.templates.TemplateManager(tmppath)

    create_template_file(
        template_manager.custom_template_path / 'story' / 'full.j2',
        'Overriden story template content.',
    )
    create_template_file(
        template_manager.custom_template_path / 'plan' / 'custom_plan.j2', 'Plan template content.'
    )

    assert template_manager.templates['story']['base'] is not None
    assert template_manager.templates['story']['full'] is not None
    assert (
        template_manager.templates['story']['full'].read_text()
        == 'Overriden story template content.'
    )

    assert template_manager.templates['plan']['base'] is not None
    assert template_manager.templates['plan']['custom_plan'] is not None
    assert (
        template_manager.templates['plan']['custom_plan'].read_text() == 'Plan template content.'
    )


def test_render_template_file(tmppath: Path) -> None:
    template_manager = tmt.templates.TemplateManager(tmppath)

    static_story_path = template_manager.custom_template_path / 'story' / 'static.j2'
    variable_story_path = template_manager.custom_template_path / 'story' / 'with_variable.j2'

    static_story_path.write_text('Story template content.')
    variable_story_path.write_text('Plan template content with variable: {{ variable }}.')

    variables = {'variable': 'value'}
    static_content = template_manager.render_file(static_story_path)
    variable_content = template_manager.render_file(variable_story_path, **variables)

    assert static_content == 'Story template content.\n'
    assert variable_content == 'Plan template content with variable: value.\n'



import datetime
import os
import re
from typing import TYPE_CHECKING, Any, Optional, overload

import requests
import urllib3

import tmt.hardware
import tmt.log
import tmt.steps.report
import tmt.utils
import tmt.utils.templates
from tmt.container import container, field
from tmt.result import ResultOutcome
from tmt.utils import (
    ActionType,
    catch_warnings_safe,
    format_timestamp,
    sanitize_string,
    yaml_to_dict,
)

if TYPE_CHECKING:
    from tmt._compat.typing import TypeAlias
    from tmt.hardware import Size

JSON: 'TypeAlias' = Any
DEFAULT_LOG_SIZE_LIMIT: 'Size' = tmt.hardware.UNITS('1 MB')
DEFAULT_TRACEBACK_SIZE_LIMIT: 'Size' = tmt.hardware.UNITS('50 kB')


def _flag_env_to_default(option: str, default: bool) -> bool:
    env_var = 'TMT_PLUGIN_REPORT_REPORTPORTAL_' + option.upper()
    if env_var not in os.environ:
        return default
    return bool(os.getenv(env_var) == '1')


@overload
def _str_env_to_default(option: str, default: None) -> Optional[str]:
    pass


@overload
def _str_env_to_default(option: str, default: str) -> str:
    pass


def _str_env_to_default(option: str, default: Optional[str]) -> Optional[str]:
    env_var = 'TMT_PLUGIN_REPORT_REPORTPORTAL_' + option.upper()
    if env_var not in os.environ or os.getenv(env_var) is None:
        return default
    return str(os.getenv(env_var))


def _size_env_to_default(option: str, default: 'Size') -> 'Size':
    return tmt.hardware.UNITS(_str_env_to_default(option, str(default)))


@container
class LogFilterSettings:
    size: 'Size' = DEFAULT_LOG_SIZE_LIMIT
    is_traceback: bool = False


def _filter_invalid_chars(data: str, settings: LogFilterSettings) -> str:
    """Filter out invalid unicode characters"""
    return sanitize_string(data)


def _filter_log_per_size(data: str, settings: LogFilterSettings) -> str:
    size = tmt.hardware.UNITS(f'{len(data)} bytes')
    if size > settings.size:
        if settings.is_traceback:
            variable = "TMT_PLUGIN_REPORT_REPORTPORTAL_TRACEBACK_SIZE_LIMIT"
            option = "--traceback-size-limit"
        else:
            variable = "TMT_PLUGIN_REPORT_REPORTPORTAL_LOG_SIZE_LIMIT"
            option = "--log-size-limit"
        header = (
            f"WARNING: Uploaded log has been truncated because its size {size} "
            f"exceeds tmt reportportal plugin limit of {settings.size}. "
            f"The limit is controlled with {option} plugin option or "
            f"{variable} environment variable.\n\n"
        )
        return f"{header}{data[: int(settings.size.to('bytes').magnitude)]}"
    return data


_LOG_FILTERS = [
    _filter_log_per_size,
    _filter_invalid_chars,
]


def _filter_log(log: str, settings: Optional[LogFilterSettings] = None) -> str:
    settings = settings or LogFilterSettings()
    for log_filter in _LOG_FILTERS:
        log = log_filter(log, settings=settings)
    return log


@container
class ReportReportPortalData(tmt.steps.report.ReportStepData):
    url: Optional[str] = field(
        option="--url",
        metavar="URL",
        default=_str_env_to_default('url', None),
        help="The URL of the ReportPortal instance where the data should be sent to.",
    )

    token: Optional[str] = field(
        option="--token",
        metavar="TOKEN",
        default=_str_env_to_default('token', None),
        help="The token to use for upload to the ReportPortal instance (from the user profile).",
    )

    project: Optional[str] = field(
        option="--project",
        metavar="PROJECT_NAME",
        default=_str_env_to_default('project', None),
        help="Name of the project into which the results should be uploaded.",
    )

    launch: Optional[str] = field(
        option="--launch",
        metavar="LAUNCH_NAME",
        default=_str_env_to_default('launch', None),
        help="""
           Set the launch name, otherwise name of the plan is used by default.
           Should be defined with 'suite-per-plan' option or it will be named after the first plan.
           """,
    )

    launch_description: Optional[str] = field(
        option="--launch-description",
        metavar="DESCRIPTION",
        default=_str_env_to_default('launch_description', None),
        help="""
             Pass the description for ReportPortal launch with 'suite-per-plan' option
             or append the original (plan summary) with additional info.
             Appends test description with 'upload-to-launch/suite' options.
             """,
    )

    launch_per_plan: bool = field(
        option="--launch-per-plan",
        default=_flag_env_to_default('launch_per_plan', False),
        is_flag=True,
        help="Mapping launch per plan, creating one or more launches with no suite structure.",
    )

    suite_per_plan: bool = field(
        option="--suite-per-plan",
        default=_flag_env_to_default('suite_per_plan', False),
        is_flag=True,
        help="""
             Mapping suite per plan, creating one launch and continuous uploading suites into it.
             Recommended to use with 'launch' and 'launch-description' options.
             Can be used with 'upload-to-launch' option for an additional upload of new suites.
             """,
    )

    upload_to_launch: Optional[str] = field(
        option="--upload-to-launch",
        metavar="LAUNCH_ID",
        default=_str_env_to_default('upload_to_launch', None),
        help="""
           Pass the launch ID for an additional test/suite upload to an existing launch. ID can be
           found in the launch URL. Keep the launch structure with options 'launch/suite-per-plan'.
           To upload specific info into description see also 'launch-description'.
           """,
    )

    upload_to_suite: Optional[str] = field(
        option="--upload-to-suite",
        metavar="SUITE_ID",
        default=_str_env_to_default('upload_to_suite', None),
        help="""
             Pass the suite ID for an additional test upload to a suite
             within an existing launch. ID can be found in the suite URL.
             To upload specific info into description see also 'launch-description'.
             """,
    )

    launch_rerun: bool = field(
        option="--launch-rerun",
        default=_flag_env_to_default('launch_rerun', False),
        is_flag=True,
        help="""
             Rerun the last launch based on its name and unique test paths to create Retry item
             with a new version per each test. Supported in 'suite-per-plan' structure only.
             """,
    )

    defect_type: Optional[str] = field(
        option="--defect-type",
        metavar="DEFECT_NAME",
        default=_str_env_to_default('defect_type', None),
        help="""
             Pass the defect type to be used for failed test, which is defined in the project
             (e.g. 'Idle'). 'To Investigate' is used by default.
             """,
    )

    log_size_limit: 'Size' = field(
        option="--log-size-limit",
        metavar="SIZE",
        default=_size_env_to_default('log_size_limit', DEFAULT_LOG_SIZE_LIMIT),
        help=f"""
              Size limit in bytes for log upload to ReportPortal.
              The default limit is {DEFAULT_LOG_SIZE_LIMIT}.
              """,
        normalize=tmt.utils.normalize_data_amount,
        serialize=lambda limit: str(limit),
        unserialize=lambda serialized: tmt.hardware.UNITS(serialized),
    )

    traceback_size_limit: 'Size' = field(
        option="--traceback-size-limit",
        metavar="SIZE",
        default=_size_env_to_default('traceback_size_limit', DEFAULT_TRACEBACK_SIZE_LIMIT),
        help=f"""
              Size limit in bytes for traceback log upload to ReportPortal.
              The default limit is {DEFAULT_TRACEBACK_SIZE_LIMIT}.
              """,
        help_example_values=[str(DEFAULT_TRACEBACK_SIZE_LIMIT), '1MB'],
        normalize=tmt.utils.normalize_data_amount,
        serialize=lambda limit: str(limit),
        unserialize=lambda serialized: tmt.hardware.UNITS(serialized),
    )

    exclude_variables: str = field(
        option="--exclude-variables",
        metavar="PATTERN",
        default=_str_env_to_default('exclude_variables', "^TMT_.*"),
        help="""
             Regular expression for excluding environment variables
             from reporting to ReportPortal ('^TMT_.*' used by default).
             Parameters in ReportPortal get filtered out by the pattern
             to prevent overloading and to preserve the history aggregation
             for ReportPortal item if tmt id is not provided.
             """,
    )

    api_version: str = field(
        option="--api-version",
        metavar="VERSION",
        default=_str_env_to_default('api_version', "v1"),
        help="Override the default reportportal API version (v1).",
    )

    artifacts_url: Optional[str] = field(
        metavar="ARTIFACTS_URL",
        option="--artifacts-url",
        default=_str_env_to_default('artifacts_url', os.getenv('TMT_REPORT_ARTIFACTS_URL')),
        help="Link to test artifacts provided for report plugins.",
    )

    ssl_verify: bool = field(
        default=True,
        option=('--ssl-verify / --no-ssl-verify'),
        is_flag=True,
        show_default=True,
        help="Enable/disable the SSL verification for communication with ReportPortal.",
    )

    upload_subresults: bool = field(
        default=False,
        option=('--upload-subresults / --no-upload-subresults'),
        is_flag=True,
        show_default=True,
        help="Enable/disable uploading of tmt subresults into the ReportPortal.",
    )

    link_template: Optional[str] = field(
        metavar="TEMPLATE",
        option="--link-template",
        default=_str_env_to_default('link_template', None),
        help="""
             Jinja template that will be rendered for each test result and appended to the end
             of its description. The following variables are passed to the template:
             ``PLAN_NAME``, ``RESULT``.
             """,
    )

    launch_url: Optional[str] = None
    launch_uuid: Optional[str] = None
    suite_uuid: Optional[str] = None
    test_uuids: dict[int, str] = field(default_factory=dict)


@tmt.steps.provides_method("reportportal")
class ReportReportPortal(tmt.steps.report.ReportPlugin[ReportReportPortalData]):
    """
    Report test results and their subresults to a ReportPortal instance via API.

    For communication with Report Portal API is necessary to provide
    following options:

    * token for authentication
    * url of the ReportPortal instance
    * project name

    In addition to command line options it's possible to use environment
    variables:

    .. code-block:: bash

        export TMT_PLUGIN_REPORT_REPORTPORTAL_${MY_OPTION}=${MY_VALUE}

        # Boolean options are activated with value of 1:
        TMT_PLUGIN_REPORT_REPORTPORTAL_SUITE_PER_PLAN=1

    Assuming the URL and token are provided by the environment variables,
    the plan config can look like this:

    .. code-block:: yaml

        # Use ReportPortal as the default report for given plan
        report:
            how: reportportal
            project: baseosqe

        # Report context attributes for given plan
        context:
            ...

        environment:
            ...

    .. code-block:: yaml

        # Report description, contact, id and environment variables for given test
        summary: ...
        contact: ...
        id: ...
        environment:
            ...

    Where the context and environment sections must be filled with
    corresponding data in order to report context as attributes
    (arch, component, distro, trigger, compose, etc.) and
    environment variables as parameters in the Item Details.

    Other reported fmf data are summary, id, web link and contact per
    test.

    Two types of data structures are supported for reporting to ReportPortal:

    * ``launch-per-plan`` mapping (default) that results in launch-test structure.
    * ``suite-per-plan`` mapping that results in launch-suite-test structure.

    Supported report use cases:

    * Report a new run in launch-suite-test or launch-test structure
    * Report an additional rerun with ``launch-rerun`` option and same launch name (-> Retry items)
      or by reusing the run and reporting with ``again`` option (-> append logs)
    * To see plan progress, discover and report an empty (IDLE) run
      and reuse the run for execution and updating the report with ``again`` option
    * Report contents of a new run to an existing launch via the URL ID in three ways:
      tests to launch, suites to launch and tests to suite.

    Example:

    .. code-block:: shell

        # Enable ReportPortal report from the command line depending on the use case:

        ## Simple upload with all project, url endpoint and user token passed in command line
        tmt run --all report --how reportportal --project=baseosqe --url="https://reportportal.xxx.com" --token="abc...789"

        ## Simple upload with url and token exported in environment variable
        tmt run --all report --how reportportal --project=baseosqe

        ## Upload with project name in fmf data, filtering out parameters (environment variables)
        ## that tend to be unique and break the history aggregation
        tmt run --all report --how reportportal --exclude-variables="^(TMT|PACKIT|TESTING_FARM).*"

        ## Upload all plans as suites into one ReportPortal launch
        tmt run --all report --how reportportal --suite-per-plan --launch=Errata --launch-description="..."

        ## Rerun the launch with suite structure for the test results to be uploaded
        ## into the latest launch with the same name as a new 'Retry' tab
        ## (mapping based on unique paths)
        tmt run --all report --how reportportal --suite-per-plan --launch=Errata --launch-rerun

        ## Rerun the tmt run and append the new result logs under the previous one
        ## uploaded in ReportPortal (precise mapping)
        tmt run --id run-012 --all report --how reportportal --again

        ## Additional upload of new suites into given launch with suite structure
        tmt run --all report --how reportportal --suite-per-plan --upload-to-launch=4321

        ## Additional upload of new tests into given launch with non-suite structure
        tmt run --all report --how reportportal --launch-per-plan --upload-to-launch=1234

        ## Additional upload of new tests into given suite
        tmt run --all report --how reportportal --upload-to-suite=123456

        ## Upload Idle tests, then execute it and add result logs into prepared empty tests
        tmt run discover report --how reportportal --defect-type=Idle
        tmt run --last --all report --how reportportal --again
    """  # noqa: E501

    _data_class = ReportReportPortalData

    TMT_TO_RP_RESULT_STATUS = {
        ResultOutcome.PASS: "PASSED",
        ResultOutcome.FAIL: "FAILED",
        ResultOutcome.INFO: "SKIPPED",
        ResultOutcome.WARN: "FAILED",
        ResultOutcome.ERROR: "FAILED",
        ResultOutcome.SKIP: "SKIPPED",
    }

    def handle_response(self, response: requests.Response) -> None:
        """
        Check the endpoint response and raise an exception if needed
        """

        self.debug("Response code from the endpoint", response.status_code)
        self.debug("Message from the endpoint", response.text)

        if not response.ok:
            raise tmt.utils.ReportError(
                f"Received non-ok status code {response.status_code} "
                f"from ReportPortal: {response.text}"
            )

    def check_options(self) -> None:
        """
        Check options for known troublesome combinations
        """

        if self.data.launch_per_plan and self.data.suite_per_plan:
            self.warn(
                "The options '--launch-per-plan' and '--suite-per-plan' are mutually exclusive. "
                "Default option '--launch-per-plan' is used."
            )
            self.data.suite_per_plan = False

        if self.data.launch_rerun and (self.data.upload_to_launch or self.data.upload_to_suite):
            self.warn(
                "Unexpected option combination: "
                "'--launch-rerun' is ignored when uploading additional tests."
            )

        if not self.data.suite_per_plan and self.data.launch_rerun:
            self.warn(
                "Unexpected option combination: '--launch-rerun' "
                "may cause an unexpected behaviour with launch-per-plan structure"
            )

    @property
    def datetime(self) -> str:
        # Use the same format of timestramp as tmt does
        return format_timestamp(datetime.datetime.now(datetime.timezone.utc))

    @property
    def headers(self) -> dict[str, str]:
        return {
            "Authorization": f"Bearer {self.data.token}",
            "Accept": "*/*",
            "Content-Type": "application/json",
        }

    @property
    def url(self) -> str:
        return f"{self.data.url}/api/{self.data.api_version}/{self.data.project}"

    def construct_launch_attributes(
        self, suite_per_plan: bool, attributes: list[dict[str, str]]
    ) -> list[dict[str, str]]:
        if not suite_per_plan or not self.step.plan.my_run:
            return attributes.copy()

        # Get common attributes across the plans
        merged_plans = [
            {key: value[0] for key, value in plan._fmf_context.items()}
            for plan in self.step.plan.my_run.plans
        ]
        result_dict = merged_plans[0]
        for current_plan in merged_plans[1:]:
            tmp_dict = {}
            for key, value in current_plan.items():
                if key in result_dict and result_dict[key] == value:
                    tmp_dict[key] = value
            result_dict = tmp_dict
        return [{'key': key, 'value': value} for key, value in result_dict.items()]

    def get_defect_type_locator(
        self, session: requests.Session, defect_type: Optional[str]
    ) -> str:
        if not defect_type:
            return "ti001"

        response = self.rp_api_get(session, "settings")
        defect_types = yaml_to_dict(response.text).get("subTypes")
        if not defect_types:
            return "ti001"

        groups_to_search = [
            'TO_INVESTIGATE',
            'NO_DEFECT',
            'SYSTEM_ISSUE',
            'AUTOMATION_BUG',
            'PRODUCT_BUG',
        ]
        for group_name in groups_to_search:
            defect_types_list = defect_types[group_name]
            dt_tmp = [
                dt['locator']
                for dt in defect_types_list
                if dt['longName'].lower() == defect_type.lower()
            ]
            dt_locator = dt_tmp[0] if dt_tmp else None
            if dt_locator:
                break
        if not dt_locator:
            raise tmt.utils.ReportError(
                f"Defect type '{defect_type}' "
                "is not be defined in the project {self.data.project}"
            )
        self.verbose("defect_type", defect_type, color="cyan", shift=1)
        return str(dt_locator)

    def rp_api_get(self, session: requests.Session, path: str) -> requests.Response:
        response = session.get(url=f"{self.url}/{path}", headers=self.headers)
        self.handle_response(response)
        return response

    def rp_api_post(self, session: requests.Session, path: str, json: JSON) -> requests.Response:
        response = session.post(url=f"{self.url}/{path}", headers=self.headers, json=json)
        self.handle_response(response)
        return response

    def rp_api_put(self, session: requests.Session, path: str, json: JSON) -> requests.Response:
        response = session.put(url=f"{self.url}/{path}", headers=self.headers, json=json)
        self.handle_response(response)
        return response

    def append_description(self, curr_description: str) -> str:
        """
        Extend text with the launch description (if provided)
        """

        if self.data.launch_description:
            if curr_description:
                curr_description += "<br>" + self.data.launch_description
            else:
                curr_description = self.data.launch_description
        return curr_description

    def append_link_template(
        self,
        description: str,
        link_template: str,
        result: tmt.result.Result,
    ) -> str:
        """
        Extend text with rendered link template
        """

        url = tmt.utils.templates.render_template(
            link_template,
            PLAN_NAME=self.step.plan.pathless_safe_name,
            RESULT=result,
        )
        return f'{description}<br>{url}' if description else url

    def upload_result_logs(
        self,
        result: tmt.result.BaseResult,
        session: requests.Session,
        item_uuid: str,
        launch_uuid: str,
        timestamp: str,
        write_out_failures: bool = True,
    ) -> None:
        """
        Upload all result log files into the ReportPortal instance
        """

        for index, log_path in enumerate(result.log):
            try:
                log = self.step.plan.execute.read(log_path)
            except tmt.utils.FileError:
                continue

            level = "INFO" if index == 0 else "TRACE"
            message = _filter_log(log, settings=LogFilterSettings(size=self.data.log_size_limit))

            # Upload log
            self.rp_api_post(
                session=session,
                path="log/entry",
                json={
                    "message": message,
                    "itemUuid": item_uuid,
                    "launchUuid": launch_uuid,
                    "level": level,
                    "time": timestamp,
                },
            )

            # Optionally write out failures only for results which implement the failures callable
            if hasattr(result, "failures") and index == 0 and write_out_failures:
                message = _filter_log(
                    result.failures(log),
                    settings=LogFilterSettings(
                        size=self.data.traceback_size_limit,
                        is_traceback=True,
                    ),
                )

                self.rp_api_post(
                    session=session,
                    path="log/entry",
                    json={
                        "message": message,
                        "itemUuid": item_uuid,
                        "launchUuid": launch_uuid,
                        "level": "ERROR",
                        "time": timestamp,
                    },
                )

    def execute_rp_import(self) -> None:
        """
        Execute the import of test, results and subresults into ReportPortal
        """

        assert self.step.plan.my_run is not None
        # Use the current datetime as a default, but this is the worst case scenario
        # and we should use timestamps from results log as much as possible.
        launch_start_time = self.datetime

        # Support for idle tests
        executed = bool(self.step.plan.execute.results())
        if executed:
            # Launch time should be the earliest start time of all plans.
            #
            # The datetime *strings* are in fact sorted here, but finding the minimum will work,
            # because the datetime in ISO format is designed to be lexicographically sortable.
            launch_start_time = min(
                [r.start_time or self.datetime for r in self.step.plan.execute.results()]
            )

        # Create launch, suites (if "--suite_per_plan") and tests;
        # or report to existing launch/suite if its id is given
        suite_per_plan = self.data.suite_per_plan
        launch_per_plan = self.data.launch_per_plan
        if not launch_per_plan and not suite_per_plan:
            launch_per_plan = True  # by default

        suite_id = self.data.upload_to_suite
        launch_id = self.data.upload_to_launch

        suite_uuid = self.data.suite_uuid
        launch_uuid = self.data.launch_uuid
        additional_upload = suite_id or launch_id or launch_uuid
        is_the_first_plan = self.step.plan == self.step.plan.my_run.plans[0]
        if not launch_uuid and suite_per_plan and not is_the_first_plan:
            rp_phases = list(self.step.plan.my_run.plans[0].report.phases(ReportReportPortal))
            if rp_phases:
                launch_uuid = rp_phases[0].data.launch_uuid

        create_test = not self.data.test_uuids
        create_suite = suite_per_plan and not (suite_uuid or suite_id)
        create_launch = not (launch_uuid or launch_id or suite_uuid or suite_id)

        launch_name = self.data.launch or self.step.plan.name
        suite_name = ""
        launch_url = ""

        launch_rerun = self.data.launch_rerun
        envar_pattern = self.data.exclude_variables or "$^"
        defect_type = self.data.defect_type or ""

        attributes = [
            {'key': key, 'value': value[0]} for key, value in self.step.plan._fmf_context.items()
        ]
        launch_attributes = self.construct_launch_attributes(suite_per_plan, attributes)

        if suite_per_plan:
            launch_description = self.data.launch_description or ""
            suite_description = self.append_description(self.step.plan.summary or "")
        else:
            launch_description = self.step.plan.summary or ""
            launch_description = self.append_description(launch_description)
            suite_description = ""

        # Check whether artifacts URL has been provided
        if self.data.artifacts_url:
            launch_description += f"<br>{self.data.artifacts_url}"
            suite_description += f"<br>{self.data.artifacts_url}"

        # Communication with RP instance
        with tmt.utils.retry_session(
            status_forcelist=(
                429,  # Too Many Requests
                500,  # Internal Server Error
                502,  # Bad Gateway
                503,  # Service Unavailable
                504,  # Gateway Timeout
            )
        ) as session:
            session.verify = self.data.ssl_verify

            if create_launch:
                # Create a launch
                self.info("launch", launch_name, color="cyan")
                response = self.rp_api_post(
                    session=session,
                    path="launch",
                    json={
                        "name": launch_name,
                        "description": launch_description,
                        "attributes": launch_attributes,
                        "startTime": launch_start_time,
                        "rerun": launch_rerun,
                    },
                )
                launch_uuid = yaml_to_dict(response.text).get("id")

            else:
                # Get the launch_uuid or info to log
                if suite_id:
                    response = self.rp_api_get(session, f"item/{suite_id}")
                    suite_uuid = yaml_to_dict(response.text).get("uuid")
                    suite_name = str(yaml_to_dict(response.text).get("name"))
                    launch_id = yaml_to_dict(response.text).get("launchId")

                if launch_id:
                    response = self.rp_api_get(session, f"launch/{launch_id}")
                    launch_uuid = yaml_to_dict(response.text).get("uuid")

            if launch_uuid and not launch_id:
                response = self.rp_api_get(session, f"launch/uuid/{launch_uuid}")
                launch_id = yaml_to_dict(response.text).get("id")

            # Print the launch info
            if not create_launch:
                launch_name = yaml_to_dict(response.text).get("name") or ""
                self.verbose("launch", launch_name, color="green")
                self.verbose("id", launch_id, "yellow", shift=1)

            assert launch_uuid is not None
            self.verbose("uuid", launch_uuid, "yellow", shift=1)
            self.data.launch_uuid = launch_uuid

            launch_url = f"{self.data.url}/ui/#{self.data.project}/launches/all/{launch_id}"

            if create_suite:
                # Create a suite
                suite_name = self.step.plan.name
                self.info("suite", suite_name, color="cyan")
                response = self.rp_api_post(
                    session=session,
                    path="item",
                    json={
                        "name": suite_name,
                        "description": suite_description,
                        "attributes": attributes,
                        "startTime": launch_start_time,
                        "launchUuid": launch_uuid,
                        "type": "suite",
                    },
                )
                suite_uuid = yaml_to_dict(response.text).get("id")
                assert suite_uuid is not None

            elif suite_name:
                self.info("suite", suite_name, color="green")
                self.verbose("id", suite_id, "yellow", shift=1)

            if suite_uuid:
                self.verbose("uuid", suite_uuid, "yellow", shift=1)
                self.data.suite_uuid = suite_uuid

            # The first test starts with the launch (at the worst case)
            test_start_time = launch_start_time

            for result, test_origin in self.step.plan.execute.results_for_tests(
                self.step.plan.discover.tests()
            ):
                test = test_origin.test if test_origin else None

                test_name = None
                test_description = ''
                test_link = None
                test_id = None
                env_vars = None

                item_attributes = attributes.copy()
                if result:
                    serial_number = result.serial_number
                    test_name = result.name

                    # Use the actual timestamp or reuse the old one if missing
                    test_start_time = result.start_time or test_start_time

                    # for guests, save their primary address
                    if result.guest.primary_address:
                        item_attributes.append(
                            {'key': 'guest_primary_address', 'value': result.guest.primary_address}
                        )
                    # for multi-host tests store also provision name and role
                    if result.guest.name != 'default-0':
                        item_attributes.append({'key': 'guest_name', 'value': result.guest.name})
                    if result.guest.role:
                        item_attributes.append({'key': 'guest_role', 'value': result.guest.role})

                # update RP item with additional attributes if test details are available
                if test:
                    serial_number = test.serial_number
                    if not test_name:
                        test_name = test.name
                    if test.author:
                        item_attributes += [
                            {'key': 'author', 'value': author} for author in test.author
                        ]
                    if test.contact:
                        item_attributes += [
                            {'key': 'contact', 'value': contact} for contact in test.contact
                        ]
                    if test.summary:
                        test_description = test.summary
                    if test.web_link():
                        test_link = test.web_link()
                    if test.id:
                        test_id = test.id

                    env_vars = [
                        {'key': key, 'value': value}
                        for key, value in test.environment.items()
                        if not re.search(envar_pattern, key)
                    ]

                if create_test:
                    if (
                        self.data.upload_to_launch and launch_per_plan
                    ) or self.data.upload_to_suite:
                        test_description = self.append_description(test_description)

                    if result and self.data.link_template:
                        test_description = self.append_link_template(
                            test_description,
                            self.data.link_template,
                            result,
                        )

                    # Create a test item
                    self.info("test", test_name, color="cyan")
                    response = self.rp_api_post(
                        session=session,
                        path=f"item{f'/{suite_uuid}' if suite_uuid else ''}",
                        json={
                            "name": test_name,
                            "description": test_description,
                            "attributes": item_attributes,
                            "parameters": env_vars,
                            "codeRef": test_link,
                            "launchUuid": launch_uuid,
                            "type": "step",
                            "testCaseId": test_id,
                            "startTime": test_start_time,
                        },
                    )

                    item_uuid = yaml_to_dict(response.text).get("id")
                    assert item_uuid is not None
                    self.verbose("uuid", item_uuid, "yellow", shift=1)
                    self.data.test_uuids[serial_number] = item_uuid
                else:
                    item_uuid = self.data.test_uuids[serial_number]

                # If the result end-time is not defined, use the latest result start-time as
                # default.
                test_end_time = test_start_time

                # Support for idle tests
                item_status = "SKIPPED"
                if result:
                    # Shift the timestamp to the end of a test.
                    test_end_time = result.end_time or test_end_time

                    item_status = self.TMT_TO_RP_RESULT_STATUS[result.result]

                    self.upload_result_logs(
                        result=result,
                        session=session,
                        item_uuid=item_uuid,
                        launch_uuid=launch_uuid,
                        timestamp=test_end_time,
                        write_out_failures=bool(item_status == "FAILED"),
                    )

                    # If upload of subresults is required, create (and *finish*) the child test
                    # item for every tmt subresult and map it under the parent test item.
                    if self.data.upload_subresults:
                        for subresult in result.subresult:
                            # Create a child item
                            self.info("sub-test", subresult.name, color="cyan", shift=1)

                            # Use the parent test start-time as a default
                            subresult_start_time = subresult.start_time or test_start_time

                            response = self.rp_api_post(
                                session=session,
                                path=f"item/{item_uuid}",
                                json={
                                    "name": subresult.name,
                                    "launchUuid": launch_uuid,
                                    "type": "step",
                                    "startTime": subresult_start_time,
                                },
                            )

                            child_item_uuid = yaml_to_dict(response.text).get("id")
                            assert child_item_uuid is not None

                            child_item_status = self.TMT_TO_RP_RESULT_STATUS[subresult.result]
                            subtest_end_time = subresult.end_time or test_end_time

                            # Upload the subtest (child) logs
                            self.upload_result_logs(
                                result=subresult,
                                session=session,
                                item_uuid=child_item_uuid,
                                launch_uuid=launch_uuid,
                                timestamp=subtest_end_time,
                                write_out_failures=bool(child_item_status == "FAILED"),
                            )

                            # Finish the child item
                            response = self.rp_api_put(
                                session=session,
                                path=f"item/{child_item_uuid}",
                                json={
                                    "launchUuid": launch_uuid,
                                    "status": child_item_status,
                                    "endTime": subtest_end_time,
                                },
                            )

                            self.verbose("uuid", child_item_uuid, "yellow", shift=2)

                # Finish the parent test item
                response = self.rp_api_put(
                    session=session,
                    path=f"item/{item_uuid}",
                    json={
                        "launchUuid": launch_uuid,
                        "endTime": test_end_time,
                        "status": item_status,
                        "issue": {"issueType": self.get_defect_type_locator(session, defect_type)},
                    },
                )

                # The launch ends with the last test
                launch_end_time = test_end_time

            if create_suite:
                # Finish the test suite
                response = self.rp_api_put(
                    session=session,
                    path=f"item{f'/{suite_uuid}' if suite_uuid else ''}",
                    json={"launchUuid": launch_uuid, "endTime": launch_end_time},
                )

            is_the_last_plan = self.step.plan == self.step.plan.my_run.plans[-1]
            if is_the_last_plan:
                self.data.defect_type = None

            if (
                launch_per_plan or (suite_per_plan and is_the_last_plan)
            ) and not additional_upload:
                # Finish the launch
                response = self.rp_api_put(
                    session=session,
                    path=f"launch/{launch_uuid}/finish",
                    json={"endTime": launch_end_time},
                )
                launch_url = str(yaml_to_dict(response.text).get("link"))

            assert launch_url is not None
            self.info("url", launch_url, "magenta")
            self.data.launch_url = launch_url

    def go(self, *, logger: Optional[tmt.log.Logger] = None) -> None:
        """
        Report test results to the endpoint

        Create a ReportPortal launch and its test items,
        fill it with all parts needed and report the logs.
        """

        super().go(logger=logger)

        if not self.data.url:
            raise tmt.utils.ReportError("No ReportPortal endpoint url provided.")
        self.data.url = self.data.url.rstrip("/")

        if not self.data.project:
            raise tmt.utils.ReportError("No ReportPortal project provided.")

        if not self.data.token:
            raise tmt.utils.ReportError("No ReportPortal token provided.")

        if not self.step.plan.my_run:
            raise tmt.utils.ReportError("No run data available.")

        self.check_options()

        # If SSL verification is disabled, do not print warnings with urllib3
        warning_filter_action: ActionType = 'default'
        if not self.data.ssl_verify:
            warning_filter_action = 'ignore'
            self.warn(
                "SSL verification is disabled for all requests being made to ReportPortal "
                f"instance ({self.data.url})."
            )

        with catch_warnings_safe(
            action=warning_filter_action, category=urllib3.exceptions.InsecureRequestWarning
        ):
            self.execute_rp_import()



import logging
from typing import Optional

import _pytest.capture
import _pytest.logging
import pytest

import tmt.utils
from tmt.log import (
    DebugLevelFilter,
    Logger,
    LogRecordDetails,
    QuietnessFilter,
    Topic,
    TopicFilter,
    VerbosityLevelFilter,
    indent,
    render_labels,
)

from . import assert_log, assert_not_log


def _exercise_logger(
    caplog: _pytest.logging.LogCaptureFixture,
    capsys: _pytest.capture.CaptureFixture[str],
    logger: Logger,
    indent_by: str = '',
    labels: Optional[list[str]] = None,
    reset: bool = True,
) -> None:
    labels = labels or []

    prefix = (
        tmt.utils.remove_color(render_labels(labels)) + indent_by + ' ' if labels else indent_by
    )

    if reset:
        caplog.clear()

    logger.print('this is printed')
    logger.debug('this is a debug message')
    logger.verbose('this is a verbose message')
    logger.info('this is just an info')
    logger.warning('this is a warning')
    logger.fail('this is a failure')

    captured = capsys.readouterr()

    assert_not_log(
        caplog,
        message=f'{prefix}this is printed',
        details_key='this is printed',
        details_logger_labels=labels,
        levelno=logging.INFO,
    )
    assert tmt.utils.remove_color(captured.out) == f'{prefix}this is printed\n'
    assert_log(
        caplog,
        message=f'{prefix}this is a debug message',
        details_key='this is a debug message',
        details_logger_labels=labels,
        levelno=logging.DEBUG,
    )
    assert_log(
        caplog,
        message=f'{prefix}this is a verbose message',
        details_key='this is a verbose message',
        details_logger_labels=labels,
        levelno=logging.INFO,
    )
    assert_log(
        caplog,
        message=f'{prefix}this is just an info',
        details_key='this is just an info',
        details_logger_labels=labels,
        levelno=logging.INFO,
    )
    assert_log(
        caplog,
        message=f'{prefix}warn: this is a warning',
        details_key='warn',
        details_value='this is a warning',
        details_logger_labels=labels,
        levelno=logging.WARNING,
    )
    assert_log(
        caplog,
        message=f'{prefix}fail: this is a failure',
        details_key='fail',
        details_value='this is a failure',
        details_logger_labels=labels,
        levelno=logging.ERROR,
    )


def test_sanity(
    caplog: _pytest.logging.LogCaptureFixture,
    capsys: _pytest.capture.CaptureFixture[str],
    root_logger: Logger,
) -> None:
    _exercise_logger(caplog, capsys, root_logger)


def test_creation(caplog: _pytest.logging.LogCaptureFixture, root_logger: Logger) -> None:
    logger = Logger.create()
    assert logger._logger.name == 'tmt'

    actual_logger = logging.Logger('3rd-party-app-logger')  # noqa: LOG001
    logger = Logger.create(actual_logger)
    assert logger._logger is actual_logger


def test_descend(
    caplog: _pytest.logging.LogCaptureFixture,
    capsys: _pytest.capture.CaptureFixture[str],
    root_logger: Logger,
) -> None:
    deeper_logger = root_logger.descend().descend().descend()

    _exercise_logger(caplog, capsys, deeper_logger, indent_by='            ')


@pytest.mark.parametrize(
    ('logger_verbosity', 'message_verbosity', 'filter_outcome'),
    [
        # (
        #   logger verbosity - corresponds to -v, -vv, -vvv CLI options,
        #   message verbosity - `level` parameter of `verbosity(...)` call,
        #   expected outcome of `VerbosityLevelFilter.filter()` - returns integer!
        # )
        (0, 1, 0),
        (1, 1, 1),
        (2, 1, 1),
        (3, 1, 1),
        (4, 1, 1),
        (0, 2, 0),
        (1, 2, 0),
        (2, 2, 1),
        (3, 2, 1),
        (4, 2, 1),
        (0, 3, 0),
        (1, 3, 0),
        (2, 3, 0),
        (3, 3, 1),
        (4, 3, 1),
        (0, 4, 0),
        (1, 4, 0),
        (2, 4, 0),
        (3, 4, 0),
        (4, 4, 1),
    ],
)
def test_verbosity_filter(
    logger_verbosity: int, message_verbosity: int, filter_outcome: int
) -> None:
    filter = VerbosityLevelFilter()

    assert (
        filter.filter(
            logging.makeLogRecord(
                {
                    'levelno': logging.INFO,
                    'details': LogRecordDetails(
                        key='dummy key',
                        logger_verbosity_level=logger_verbosity,
                        message_verbosity_level=message_verbosity,
                    ),
                }
            )
        )
        == filter_outcome
    )


@pytest.mark.parametrize(
    ('logger_debug', 'message_debug', 'filter_outcome'),
    [
        # (
        #   logger debug level - corresponds to -d, -dd, -ddd CLI options,
        #   message debug level - `level` parameter of `debug(...)` call,
        #   expected outcome of `DebugLevelFilter.filter()` - returns integer!
        # )
        (0, 1, 0),
        (1, 1, 1),
        (2, 1, 1),
        (3, 1, 1),
        (4, 1, 1),
        (0, 2, 0),
        (1, 2, 0),
        (2, 2, 1),
        (3, 2, 1),
        (4, 2, 1),
        (0, 3, 0),
        (1, 3, 0),
        (2, 3, 0),
        (3, 3, 1),
        (4, 3, 1),
        (0, 4, 0),
        (1, 4, 0),
        (2, 4, 0),
        (3, 4, 0),
        (4, 4, 1),
    ],
)
def test_debug_filter(logger_debug: int, message_debug: int, filter_outcome: int) -> None:
    filter = DebugLevelFilter()

    assert (
        filter.filter(
            logging.makeLogRecord(
                {
                    'levelno': logging.DEBUG,
                    'details': LogRecordDetails(
                        key='dummy key',
                        logger_debug_level=logger_debug,
                        message_debug_level=message_debug,
                    ),
                }
            )
        )
        == filter_outcome
    )


@pytest.mark.parametrize(
    ('levelno', 'filter_outcome'),
    [
        # (
        #   log message level,
        #   expected outcome of `QietnessFilter.filter()` - returns integer!
        # )
        (logging.DEBUG, 0),
        (logging.INFO, 0),
        (logging.WARNING, 0),
        (logging.ERROR, 1),
        (logging.CRITICAL, 1),
    ],
)
def test_quietness_filter(levelno: int, filter_outcome: int) -> None:
    filter = QuietnessFilter()

    assert filter.filter(logging.makeLogRecord({'levelno': levelno})) == filter_outcome


def test_labels(
    caplog: _pytest.logging.LogCaptureFixture,
    capsys: _pytest.capture.CaptureFixture[str],
    root_logger: Logger,
) -> None:
    _exercise_logger(caplog, capsys, root_logger, labels=[])

    root_logger.labels += ['foo']

    _exercise_logger(caplog, capsys, root_logger, labels=['foo'])

    root_logger.labels += ['bar']

    _exercise_logger(caplog, capsys, root_logger, labels=['foo', 'bar'])


def test_bootstrap_logger(
    caplog: _pytest.logging.LogCaptureFixture, capsys: _pytest.capture.CaptureFixture[str]
) -> None:
    _exercise_logger(caplog, capsys, Logger.get_bootstrap_logger())


# Helpers for the test below, to make strings slightly shorter: Rendered Labels...
RL = render_labels(["foo", "bar"])
# ... and Rendered Labels with Padding.
RLP = render_labels(["foo", "bar"]) + '   '


@pytest.mark.parametrize(
    ('key', 'value', 'color', 'level', 'labels', 'labels_padding', 'expected'),
    [
        ('dummy-key', None, None, 0, None, 0, 'dummy-key\033[0m'),
        ('dummy-key', 'dummy-value', None, 0, None, 0, 'dummy-key\033[0m: dummy-value'),
        (
            'dummy-key',
            'dummy\nmultiline\nvalue',
            None,
            0,
            None,
            0,
            'dummy-key\033[0m:\n    dummy\n    multiline\n    value',
        ),
        ('dummy-key', None, None, 2, None, 0, '        dummy-key\033[0m'),
        ('dummy-key', 'dummy-value', None, 2, None, 0, '        dummy-key\033[0m: dummy-value'),
        (
            'dummy-key',
            'dummy\nmultiline\nvalue',
            None,
            2,
            None,
            0,
            '        dummy-key\033[0m:\n            dummy\n            multiline\n            value',  # noqa: E501
        ),
        (
            'dummy-key',
            'dummy\nmultiline\nvalue',
            None,
            2,
            ['foo', 'bar'],
            0,
            f'{RL}         dummy-key\033[0m:\n'
            f'{RL}             dummy\n'
            f'{RL}             multiline\n'
            f'{RL}             value',
        ),
        (
            'dummy-key',
            'dummy\nmultiline\nvalue',
            None,
            2,
            ['foo', 'bar'],
            # Pad labels to occupy their actual length plus 3 more characters
            len(RL) + 3,
            f'{RLP}         dummy-key\033[0m:\n'
            f'{RLP}             dummy\n'
            f'{RLP}             multiline\n'
            f'{RLP}             value',
        ),
    ],
    ids=[
        'key only',
        'key and value',
        'key and multiline value',
        'key only, indented',
        'key and value, indented',
        'key and multiline value, indented',
        'key and multiline value, indented, with labels',
        'key and multiline value, indented, with labels, padded',
    ],
)
def test_indent(key, value, color, level, labels, labels_padding, expected):
    assert (
        indent(
            key,
            value=value,
            color=color,
            level=level,
            labels=labels,
            labels_padding=labels_padding,
        )
        == expected
    )


@pytest.mark.parametrize(
    ('logger_topics', 'message_topic', 'filter_outcome'),
    [
        # (
        #     logger topics,
        #     message topic,
        #   expected outcome of `QietnessFilter.filter()`
        # )
        (set(), None, True),
        (set(), Topic.KEY_NORMALIZATION, False),
        ({Topic.KEY_NORMALIZATION}, Topic.KEY_NORMALIZATION, True),
        ({Topic.KEY_NORMALIZATION}, None, True),
    ],
    ids=(
        'no logger topics, no message topic',
        'no logger topics, message has topic',
        'message for enabled topic',
        'logger topic, no message topic',
        # TODO: enable once we have more than one topic
        # 'message for disabled topic'
    ),
)
def test_topic_filter(
    logger_topics: set[Topic], message_topic: Optional[Topic], filter_outcome: bool
) -> None:
    filter = TopicFilter()

    assert (
        filter.filter(
            logging.makeLogRecord(
                {
                    'levelno': logging.INFO,
                    'details': LogRecordDetails(
                        key='dummy key', logger_topics=logger_topics, message_topic=message_topic
                    ),
                }
            )
        )
        == filter_outcome
    )



from textwrap import dedent

import pytest

from tmt.convert import extract_relevancy, filter_common_data
from tmt.utils.structured_field import StructuredField


def test_extract_relevancy_field_has_priority():
    notes = dedent("""
    relevancy:
    distro == fedora: False

    [structured-field-start]
    This is StructuredField version 1. Please, edit with care.

    [relevancy]
    distro == rhel: False

    [structured-field-end]
  """)

    field = StructuredField(notes)
    relevancy = extract_relevancy(notes, field)
    assert relevancy == "distro == rhel: False\n"


@pytest.mark.parametrize(
    ('expected', 'individual'),
    [
        (
            {'common': {1: 1, 2: 2}, 'individual': [{}, {}]},
            [{1: 1, 2: 2}, {1: 1, 2: 2}],
        ),
        (
            {'common': {1: 1}, 'individual': [{2: 2}, {3: 3}]},
            [{1: 1, 2: 2}, {1: 1, 3: 3}],
        ),
        (
            {'common': {1: 1}, 'individual': [{2: 2}, {3: 3}, {}]},
            [{1: 1, 2: 2}, {1: 1, 3: 3}, {1: 1}],
        ),
        (
            {'common': {}, 'individual': [{1: 1, 2: 2}, {3: 3}]},
            [{1: 1, 2: 2}, {3: 3}],
        ),
        (
            {'common': {}, 'individual': [{1: 1, 2: 2}, {3: 3}, {3: 3, 4: 4}]},
            [{1: 1, 2: 2}, {3: 3}, {3: 3, 4: 4}],
        ),
    ],
    ids=(
        'everything-common',
        'some-matches',
        'three-way-some-matches',
        'no-matches',
        'three-way-no-matches',
    ),
)
def test_filter_common_data(expected, individual):
    common = {}
    filter_common_data(common, individual)
    assert common == expected['common']
    assert individual == expected['individual']



import logging
import operator
import re
from collections.abc import Iterable
from typing import Any, Callable

import _pytest.logging
import pytest
from pytest_container import Container

from tmt.utils import remove_color


class PatternMatching:
    def __init__(self, pattern: str, method: str) -> None:
        self.pattern = pattern
        self._compiled_pattern = re.compile(pattern)
        self.method = getattr(self._compiled_pattern, method)

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}: "{self.pattern}">'


class MATCH(PatternMatching):
    """
    Wrap a string with this class, to use it as a regular expression when matching log records.

    :py:class:`SEARCH` applies to any place within the string, while ``MATCH`` must match from the
    beginning of the string.

    .. code-block:: python

       assert_log(message=MATCH('an exception .+ was raised'))
    """

    def __init__(self, pattern: str) -> None:
        super().__init__(pattern, 'match')


class SEARCH(PatternMatching):
    """
    Wrap a string with this class, to use it as a regular expression when searching for log
    records.

    ``SEARCH`` applies to any place within the string, while :py:class:`MATCH` must match from the
    beginning of the string.

    .. code-block:: python

       assert_log(message=SEARCH('an exception .+ was raised'))
    """

    def __init__(self, pattern: str) -> None:
        super().__init__(pattern, 'search')


def _assert_log(
    caplog: _pytest.logging.LogCaptureFixture,
    evaluator: Callable[[Iterable[Any]], bool] = any,
    remove_colors: bool = False,
    not_present: bool = False,
    **tests: Any,
) -> None:
    """
    Assert log contains a record - logged message - with given properties. Those are specified as
    keyword parameters: :py:class:`logging.LogRecords` properties are allowed names, parameter
    values are the expected values.

    .. code-block:: python

       assert_log(message='everything went well', levelno=logging.INFO)
       assert_log(message='things broke down', levelno=logging.ERROR)
       assert_log(message=MATCH('user .+ logged in'), levelno=logging.INFO)

    :param caplog: Pytest's `caplog` fixture.
    :param evaluator: a callable reducing a given list of booleans into a single boolean. It is
        used to evaluate whether the search for matching record was successful: each record is
        tested, and results of these per-record tests are passed to `evaluator` for the final
        decision.
    """

    # We are given field_name=expected_value pairs, but we also want to be open to other binary
    # operators, like "field_name matches pattern". To protect the actual matching from aspects of
    # different possible operators, we will convert the "tests" into basic building blocks: a
    # field name, a callable accepting two parameters, and the given (expected) value. With these,
    # we can reduce the matching into functions calls without worrying what functions we work with.

    operators: list[tuple[Callable[[Any], Any], str, Callable[[Any, Any], bool], Any]] = []

    for field_name, expected_value in tests.items():
        if field_name == 'message' and remove_colors:

            def field_getter(record, name):
                return remove_color(getattr(record, name))

        elif field_name.startswith('details_'):
            field_name = field_name.replace('details_', '')

            def field_getter(record, name):
                return getattr(record.details, name, None)

        elif field_name == 'message':

            def field_getter(record, name):
                return remove_color(getattr(record, name))

        else:

            def field_getter(record, name):
                return getattr(record, name)

        # Special case: if the expected value is a pattern matching instance, it represents a
        # regular expression. We don't modify the field name and "expected" value, but the
        # function will be a custom lambda calling proper `re` method.
        if isinstance(expected_value, PatternMatching):
            operators.append(
                (field_getter, field_name, lambda a, b: a.method(b) is not None, expected_value)
            )

            continue

        # Python's `operator` package offers operators - `==` or `!=` - in a form of functions,
        # which is exactly what we need here, so we don't have to build our own
        # `lambda a, b: a == b`. We might use more than just `eq` in the future, so let's start
        # with `operator` right away.

        operators.append((field_getter, field_name, operator.eq, expected_value))

    # Given a logging record, apply all field/operator/value triplets, and make sure all match the
    # actual record properties.
    def _cmp(record: logging.LogRecord) -> bool:
        return all(
            op(expected_value, field_getter(record, field_name))
            for field_getter, field_name, op, expected_value in operators
        )

    # Final step: apply our "make sure field/operator/value triplets match given record" to each
    # and every record, and reduce per-record results into a single answer. By default, `any` is
    # used which means that any record matching all field/operator/value triples yield the final
    # "yes, such a record exists".
    outcome = evaluator([_cmp(record) for record in caplog.records])

    def _report(message: str) -> None:
        formatted_fields = [f'    {field} == {value}' for field, value in tests.items()]

        for record in caplog.records:
            for field_getter, field_name, op, expected_value in operators:
                print(
                    f'field={field_name}',
                    f'current=>>>{field_getter(record, field_name)}<<<',
                    f'expected=>>>{expected_value}<<<',
                    f'comparison={op(expected_value, field_getter(record, field_name))}',
                )

        pytest.fail(f"""
{message}:

{chr(10).join(formatted_fields)}
""")

    # Cannot find log record with these properties

    if not_present is False and not outcome:
        _report('Could not find log record with these properties when it should exist')

    elif not_present is True and outcome:
        _report('Found log record with these properties when it should not exist')


def assert_log(
    caplog: _pytest.logging.LogCaptureFixture,
    evaluator: Callable[[Iterable[Any]], bool] = any,
    remove_colors: bool = False,
    **tests: Any,
) -> None:
    _assert_log(
        caplog, evaluator=evaluator, remove_colors=remove_colors, not_present=False, **tests
    )


def assert_not_log(
    caplog: _pytest.logging.LogCaptureFixture,
    evaluator: Callable[[Iterable[Any]], bool] = any,
    remove_colors: bool = False,
    **tests: Any,
) -> None:
    _assert_log(
        caplog, evaluator=evaluator, remove_colors=remove_colors, not_present=True, **tests
    )


# Local images created via `make images/test`, reference to local registry
CONTAINER_FEDORA_RAWHIDE = Container(
    url='containers-storage:localhost/tmt/container/test/fedora/rawhide/upstream:latest'
)
CONTAINER_FEDORA_41 = Container(
    url='containers-storage:localhost/tmt/container/test/fedora/41/upstream:latest'
)
CONTAINER_FEDORA_40 = Container(
    url='containers-storage:localhost/tmt/container/test/fedora/40/upstream:latest'
)
CONTAINER_CENTOS_STREAM_10 = Container(
    url='containers-storage:localhost/tmt/container/test/centos/stream10/upstream:latest'
)
CONTAINER_CENTOS_STREAM_9 = Container(
    url='containers-storage:localhost/tmt/container/test/centos/stream9/upstream:latest'
)
CONTAINER_CENTOS_7 = Container(
    url='containers-storage:localhost/tmt/container/test/centos/7/upstream:latest'
)
CONTAINER_UBI_8 = Container(
    url='containers-storage:localhost/tmt/container/test/ubi/8/upstream:latest'
)
CONTAINER_UBUNTU_2204 = Container(
    url='containers-storage:localhost/tmt/container/test/ubuntu/22.04/upstream:latest'
)
CONTAINER_DEBIAN_127 = Container(
    url='containers-storage:localhost/tmt/container/test/debian/12.7/upstream:latest'
)
CONTAINER_FEDORA_COREOS = Container(
    url='containers-storage:localhost/tmt/container/test/fedora/coreos:stable'
)
CONTAINER_FEDORA_COREOS_OSTREE = Container(
    url='containers-storage:localhost/tmt/container/test/fedora/coreos/ostree:stable'
)
CONTAINER_ALPINE = Container(url='containers-storage:localhost/tmt/container/test/alpine:latest')


TEST_CONTAINERS: dict[str, Container] = {
    _container_attribute: _container
    for _container_attribute, _container in locals().items()
    if _container_attribute.startswith('CONTAINER_')
}



import re
from collections import OrderedDict
from collections.abc import Iterator
from typing import Any, Optional, Union

import fmf.utils

from tmt.utils import StructuredFieldError, format_value, pure_ascii

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#  StructuredField
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

log = fmf.utils.Logging('tmt').logger

SFSectionValueType = Union[str, list[str]]


class StructuredField:
    """
    Handling multiple text data in a single text field

    The StructuredField allows you to easily store and extract several
    sections of text data to/from a single text field. The sections are
    separated by section names in square brackets and can be hosted in
    other text as well.

    The section names have to be provided on a separate line and there
    must be no leading/trailing white space before/after the brackets.
    The StructuredField supports two versions of the format:

    Version 0: Simple, concise, useful when neither the surrounding text
    or the section data can contain lines which could resemble section
    names. Here's an example of a simple StructuredField:

    .. code-block:: ini

        Note written by human.

        [section-one]
        Section one content.

        [section-two]
        Section two content.

        [section-three]
        Section three content.

        [end]

        Another note written by human.

    Version 1: Includes unique header to prevent collisions with the
    surrounding text and escapes any section-like lines in the content:

    .. code-block:: ini

        Note written by human.

        [structured-field-start]
        This is StructuredField version 1. Please, edit with care.

        [section-one]
        Section one content.

        [section-two]
        Section two content.
        [structured-field-escape][something-resembling-section-name]

        [section-three]
        Section three content.

        [structured-field-end]

        Another note written by human.

    Note that an additional empty line is added at the end of each
    section to improve the readability. This line is not considered
    to be part of the section content.

    Besides handling the whole section content it's also possible to
    store several key-value pairs in a single section, similarly as in
    the ini config format:

    .. code-block:: ini

        [section]
        key1 = value1
        key2 = value2
        key3 = value3

    Provide the key name as the optional argument 'item' when accessing
    these single-line items. Note that the section cannot contain both
    plain text data and key-value pairs.

    .. code-block:: python

        field = qe.StructuredField()
        field.set("project", "Project Name")
        field.set("details", "somebody", "owner")
        field.set("details", "2013-05-27", "started")
        field.set("description", "This is a description.\\n"
                "It spans across multiple lines.\\n")
        print field.save()

            [structured-field-start]
            This is StructuredField version 1. Please, edit with care.

            [project]
            Project Name

            [details]
            owner = somebody
            started = 2013-05-27

            [description]
            This is a description.
            It spans across multiple lines.

            [structured-field-end]

        field.version(0)
        print field.save()

            [project]
            Project Name

            [details]
            owner = somebody
            started = 2013-05-27

            [description]
            This is a description.
            It spans across multiple lines.

            [end]

    Multiple values for the same key are supported as well. Enable this
    feature with 'multi=True' when initializing the structured field.
    If multiple values are present their list will be returned instead
    of a single string. Similarly use list for setting multiple values:

    .. code-block:: python

        field = qe.StructuredField(multi=True)
        requirements = ['hypervisor=', 'labcontroller=lab.example.com']
        field.set("hardware", requirements, "hostrequire")
        print field.save()

            [structured-field-start]
            This is StructuredField version 1. Please, edit with care.

            [hardware]
            hostrequire = hypervisor=
            hostrequire = labcontroller=lab.example.com

            [structured-field-end]

        print field.get("hardware", "hostrequire")

            ['hypervisor=', 'labcontroller=lab.example.com']

    """

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    #  StructuredField Special
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    def __init__(self, text: Optional[str] = None, version: int = 1, multi: bool = False) -> None:
        """
        Initialize the structured field
        """

        self.version(version)
        self._header: str = ""
        self._footer: str = ""
        # Sections are internally stored in their serialized form, i.e. as
        # strings.
        self._sections: dict[str, str] = {}
        self._order: list[str] = []
        self._multi = multi
        if text is not None:
            self.load(text)

    def __iter__(self) -> Iterator[str]:
        """
        By default iterate through all available section names
        """

        yield from self._order

    def __nonzero__(self) -> bool:
        """
        True when any section is defined
        """

        return len(self._order) > 0

    __bool__ = __nonzero__

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    #  StructuredField Private
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    def _load_version_zero(self, text: str) -> None:
        """
        Load version 0 format
        """

        # Attempt to split the text according to the section tag
        section = re.compile(r"\n?^\[([^\]]+)\]\n", re.MULTILINE)
        parts = section.split(text)
        # If just one part ---> no sections present, just plain text
        if len(parts) == 1:
            self._header = parts[0]
            return
        # Pick header & footer, make sure [end] tag is present
        self._header = parts[0]
        self._footer = re.sub("^\n", "", parts[-1])
        if parts[-2] != "end":
            raise StructuredFieldError("No [end] section tag found")
        # Convert to dictionary and save the order
        keys = parts[1:-2:2]
        values = parts[2:-2:2]
        for key, value in zip(keys, values):
            self.set(key, value)

    def _load(self, text: str) -> None:
        """
        Load version 1+ format
        """

        # The text must exactly match the format
        format = re.compile(
            r"(.*)^\[structured-field-start\][ \t]*\n"
            r"(.*)\n\[structured-field-end\][ \t]*\n(.*)",
            re.DOTALL + re.MULTILINE,
        )
        # No match ---> plain text or broken structured field
        matched = format.search(text)
        if not matched:
            if "[structured-field" in text:
                raise StructuredFieldError("StructuredField parse error")
            self._header = text
            log.debug("StructuredField not found, treating as a plain text")
            return
        # Save header & footer (remove trailing new lines)
        self._header = re.sub("\n\n$", "\n", matched.groups()[0])
        if self._header:
            log.debug(f"Parsed header:\n{self._header}")
        self._footer = re.sub("^\n", "", matched.groups()[2])
        if self._footer:
            log.debug(f"Parsed footer:\n{self._footer}")
        # Split the content on the section names
        section = re.compile(r"\n\[([^\]]+)\][ \t]*\n", re.MULTILINE)
        parts = section.split(matched.groups()[1])
        # Detect the version
        version_match = re.search(r"version (\d+)", parts[0])
        if not version_match:
            log.error(parts[0])
            raise StructuredFieldError("Unable to detect StructuredField version")
        self.version(int(version_match.groups()[0]))
        log.debug(f"Detected StructuredField version {self.version()}")
        # Convert to dictionary, remove escapes and save the order
        keys = parts[1::2]
        escape = re.compile(r"^\[structured-field-escape\]", re.MULTILINE)
        values = [escape.sub("", value) for value in parts[2::2]]
        for key, value in zip(keys, values):
            self.set(key, value)
        log.debug(f"Parsed sections:\n{format_value(self._sections)}")

    def _save_version_zero(self) -> str:
        """
        Save version 0 format
        """

        result = []
        if self._header:
            result.append(self._header)
        for section, content in self.iterate():
            result.append(f"[{section}]\n{content}")
        if self:
            result.append("[end]\n")
        if self._footer:
            result.append(self._footer)
        return "\n".join(result)

    def _save(self) -> str:
        """
        Save version 1+ format
        """

        result = []
        # Regular expression for escaping section-like lines
        escape = re.compile(r"^(\[.+\])$", re.MULTILINE)
        # Header
        if self._header:
            result.append(self._header)
        # Sections
        if self:
            result.append(
                "[structured-field-start]\n"
                f"This is StructuredField version {self._version}. "
                "Please, edit with care.\n"
            )
            for section, content in self.iterate():
                result.append(
                    "[{}]\n{}".format(section, escape.sub("[structured-field-escape]\\1", content))
                )
            result.append("[structured-field-end]\n")
        # Footer
        if self._footer:
            result.append(self._footer)
        return "\n".join(result)

    def _read_section(self, content: str) -> dict[str, SFSectionValueType]:
        """
        Parse config section and return ordered dictionary
        """

        dictionary: dict[str, SFSectionValueType] = OrderedDict()
        for line in content.split("\n"):
            # Remove comments and skip empty lines
            line = re.sub("#.*", "", line)
            if re.match(r"^\s*$", line):
                continue
            # Parse key and value
            matched = re.search("([^=]+)=(.*)", line)
            if not matched:
                raise StructuredFieldError(f"Invalid key/value line: {line}")
            key = matched.groups()[0].strip()
            value = matched.groups()[1].strip()
            # Handle multiple values if enabled
            if key in dictionary and self._multi:
                stored_value = dictionary[key]
                if isinstance(stored_value, list):
                    stored_value.append(value)
                else:
                    dictionary[key] = [stored_value, value]
            else:
                dictionary[key] = value
        return dictionary

    def _write_section(self, dictionary: dict[str, SFSectionValueType]) -> str:
        """
        Convert dictionary into a config section format
        """

        section = ""
        for key, value in dictionary.items():
            if isinstance(value, list):
                for item in value:
                    section += f"{key} = {item}\n"
            else:
                section += f"{key} = {value}\n"
        return section

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    #  StructuredField Methods
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    def iterate(self) -> Iterator[tuple[str, str]]:
        """
        Return (section, content) tuples for all sections
        """

        for section in self:
            yield section, self._sections[section]

    def version(self, version: Optional[int] = None) -> int:
        """
        Get or set the StructuredField version
        """

        if version is not None:
            if version in [0, 1]:
                self._version = version
            else:
                raise StructuredFieldError(f"Bad StructuredField version: {version}")
        return self._version

    def load(self, text: str, version: Optional[int] = None) -> None:
        """
        Load the StructuredField from a string
        """

        if version is not None:
            self.version(version)
        # Make sure we got a text, convert from bytes if necessary
        if isinstance(text, bytes):
            text = text.decode("utf8")
        if not isinstance(text, str):
            raise StructuredFieldError("Invalid StructuredField, expecting string")
        # Remove possible carriage returns
        text = re.sub("\r\n", "\n", text)
        # Make sure the text has a new line at the end
        if text and text[-1] != "\n":
            text += "\n"
        log.debug(f"Parsing StructuredField\n{text}")
        # Parse respective format version
        if self._version == 0:
            self._load_version_zero(text)
        else:
            self._load(text)

    def save(self) -> str:
        """
        Convert the StructuredField into a string
        """

        if self.version() == 0:
            return self._save_version_zero()
        return self._save()

    def header(self, content: Optional[str] = None) -> str:
        """
        Get or set the header content
        """

        if content is not None:
            self._header = content
        return self._header

    def footer(self, content: Optional[str] = None) -> str:
        """
        Get or set the footer content
        """

        if content is not None:
            self._footer = content
        return self._footer

    def sections(self) -> list[str]:
        """
        Get the list of available sections
        """

        return self._order

    def get(self, section: str, item: Optional[str] = None) -> SFSectionValueType:
        """
        Return content of given section or section item
        """

        try:
            content = self._sections[section]
        except KeyError:
            raise StructuredFieldError(f"Section [{pure_ascii(section)!r}] not found")
        # Return the whole section content
        if item is None:
            return content
        # Return only selected item from the section
        try:
            return self._read_section(content)[item]
        except KeyError:
            raise StructuredFieldError(
                f"Unable to read '{pure_ascii(item)!r}' from section '{pure_ascii(section)!r}'"
            )

    def set(self, section: str, content: Any, item: Optional[str] = None) -> None:
        """
        Update content of given section or section item
        """

        # Convert to string if necessary, keep lists untouched
        if isinstance(content, list):
            pass
        elif isinstance(content, bytes):
            content = content.decode("utf8")
        elif not isinstance(content, str):
            content = str(content)
        # Set the whole section content
        if item is None:
            # Add new line if missing
            if content and content[-1] != "\n":
                content += "\n"
            self._sections[section] = content
        # Set only selected item from the section
        else:
            try:
                current = self._sections[section]
            except KeyError:
                current = ""
            dictionary = self._read_section(current)
            dictionary[item] = content
            self._sections[section] = self._write_section(dictionary)
        # Remember the order when adding a new section
        if section not in self._order:
            self._order.append(section)

    def remove(self, section: str, item: Optional[str] = None) -> None:
        """
        Remove given section or section item
        """

        # Remove the whole section
        if item is None:
            try:
                del self._sections[section]
                del self._order[self._order.index(section)]
            except KeyError:
                raise StructuredFieldError(f"Section [{pure_ascii(section)!r}] not found")
        # Remove only selected item from the section
        else:
            try:
                dictionary = self._read_section(self._sections[section])
                del dictionary[item]
            except KeyError:
                raise StructuredFieldError(
                    f"Unable to remove '{pure_ascii(item)!r}' "
                    f"from section '{pure_ascii(section)!r}'"
                )
            self._sections[section] = self._write_section(dictionary)



import shutil
from unittest.mock import MagicMock

import pytest

import tmt
import tmt.base
import tmt.libraries.beakerlib
import tmt.utils.git
from tmt.utils import Path


@pytest.mark.web
def test_basic(root_logger):
    """
    Fetch a beakerlib library with/without providing a parent
    """

    parent = tmt.utils.Common(logger=root_logger, workdir=True)
    library_with_parent = tmt.libraries.library_factory(
        logger=root_logger,
        identifier=tmt.base.DependencySimple('library(openssl/certgen)'),
        parent=parent,
    )
    library_without_parent = tmt.libraries.library_factory(
        logger=root_logger, identifier=tmt.base.DependencySimple('library(openssl/certgen)')
    )

    for library in [library_with_parent, library_without_parent]:
        assert library.format == 'rpm'
        assert library.repo == Path('openssl')
        assert library.url == 'https://github.com/beakerlib/openssl'
        assert library.ref == 'master'  # The default branch is master
        assert (
            library.dest.resolve()
            == Path.cwd().joinpath(tmt.libraries.beakerlib.DEFAULT_DESTINATION).resolve()
        )
        shutil.rmtree(library.parent.workdir)


@pytest.mark.web
@pytest.mark.parametrize(
    ('url', 'name', 'default_branch'),
    [
        ('https://github.com/beakerlib/httpd', '/http', 'master'),
        ('https://github.com/beakerlib/example', '/file', 'main'),
    ],
)
def test_require_from_fmf(url, name, default_branch, root_logger):
    """
    Fetch beakerlib library referenced by fmf identifier
    """

    library = tmt.libraries.library_factory(
        logger=root_logger, identifier=tmt.base.DependencyFmfId(url=url, name=name)
    )
    assert library.format == 'fmf'
    assert library.ref == default_branch
    assert library.url == url
    assert (
        library.dest.resolve()
        == Path.cwd().joinpath(tmt.libraries.beakerlib.DEFAULT_DESTINATION).resolve()
    )
    assert library.repo == Path(url.split('/')[-1])
    assert library.name == name
    shutil.rmtree(library.parent.workdir)


@pytest.mark.web
def test_invalid_url_conflict(root_logger):
    """
    Saner check if url mismatched for translated library
    """

    parent = tmt.utils.Common(logger=root_logger, workdir=True)
    # Fetch to cache 'tmt' repo
    tmt.libraries.library_factory(
        logger=root_logger,
        identifier=tmt.base.DependencyFmfId(
            url='https://github.com/teemtee/tmt',
            name='/',
            path=Path('/tests/libraries/local/data'),
        ),
        parent=parent,
    )
    # Library 'tmt' repo is already fetched from different git,
    # however upstream (gh.com/beakerlib/tmt) repo does not exist,
    # so there can't be "already fetched" error
    with pytest.raises(tmt.libraries.LibraryError):
        tmt.libraries.library_factory(
            logger=root_logger, identifier='library(tmt/foo)', parent=parent
        )
    shutil.rmtree(parent.workdir)


@pytest.mark.web
def test_dependencies(root_logger):
    """
    Check requires for possible libraries
    """

    parent = tmt.utils.Common(logger=root_logger, workdir=True)
    requires, recommends, libraries = tmt.libraries.dependencies(
        original_require=[
            tmt.base.DependencySimple('library(httpd/http)'),
            tmt.base.DependencySimple('wget'),
        ],
        original_recommend=[tmt.base.DependencySimple('forest')],
        parent=parent,
        logger=root_logger,
    )
    # Check for correct requires and recommends
    for require in ['httpd', 'lsof', 'mod_ssl']:
        assert require in requires
        assert require in libraries[0].require
    assert 'openssl' in libraries[2].require
    assert 'forest' in recommends
    assert 'wget' in requires
    # Library require should be in httpd requires but not in the final result
    assert 'library(openssl/certgen)' in libraries[0].require
    assert 'library(openssl/certgen)' not in requires
    # Check library attributes for sane values
    assert libraries[0].repo == Path('httpd')
    assert libraries[0].name == '/http'
    assert libraries[0].url == 'https://github.com/beakerlib/httpd'
    assert libraries[0].ref == 'master'  # The default branch is master
    assert (
        libraries[0].dest.resolve()
        == Path.cwd().joinpath(tmt.libraries.beakerlib.DEFAULT_DESTINATION).resolve()
    )
    assert libraries[1].repo == Path('openssl')
    assert libraries[1].name == '/certgen'
    shutil.rmtree(parent.workdir)


@pytest.mark.web
def test_mark_nonexistent_url(root_logger, monkeypatch):
    """
    Check url existence just one time
    """

    parent = tmt.utils.Common(logger=root_logger, workdir=True)
    identifier = tmt.base.DependencyFmfId(
        url='https://github.com/beakerlib/THISDOESNTEXIST',
        name='/',
    )
    with pytest.raises(tmt.utils.GeneralError):
        tmt.libraries.beakerlib.BeakerLib(
            logger=root_logger, identifier=identifier, parent=parent
        ).fetch()
    # Second time there shouldn't be an attempt to clone...
    monkeypatch.setattr(
        "tmt.utils.git.git_clone", MagicMock(side_effect=RuntimeError('Should not be called'))
    )
    with pytest.raises(tmt.utils.GeneralError):
        tmt.libraries.beakerlib.BeakerLib(
            logger=root_logger, identifier=identifier, parent=parent
        ).fetch()
    shutil.rmtree(parent.workdir)



from typing import Optional

import tmt.log
import tmt.result
import tmt.steps.execute
import tmt.utils
from tmt.frameworks import TestFramework, provides_framework
from tmt.result import ResultOutcome
from tmt.steps.execute import TEST_OUTPUT_FILENAME, TestInvocation


@provides_framework('shell')
class Shell(TestFramework):
    @classmethod
    def get_test_command(
        cls, invocation: 'TestInvocation', logger: tmt.log.Logger
    ) -> tmt.utils.ShellScript:
        # Use default options for shell tests
        return tmt.utils.ShellScript(f"{tmt.utils.SHELL_OPTIONS}; {invocation.test.test}")

    @classmethod
    def _process_results_reduce(
        cls, invocation: TestInvocation, results: list['tmt.result.RawResult']
    ) -> list['tmt.result.Result']:
        """
        Reduce given results to one outcome.

        This is the default behavior applied to given test results: all results will be reduced to
        the worst outcome possible.

        Also, convert the ``results`` into the :py:class:`SubResult` instances, and return them as
        part of returned :py:class:`Result` instance.

        :param invocation: test invocation to which the results belong to.
        :param results: results to reduce and save as tmt subresults.
        :returns: list of results.
        """

        # The worst result outcome we can find among loaded results...
        original_outcome: Optional[ResultOutcome] = None
        # ... and the actual outcome we decided is the best representing
        # the results.
        # The original one may be left unset - malformed results file,
        # for example, provides no usable original outcome.
        actual_outcome: ResultOutcome
        note: Optional[str] = None

        try:
            outcomes = [result.result for result in results]
        except tmt.utils.SpecificationError as exc:
            actual_outcome = ResultOutcome.ERROR
            note = exc.message
        else:
            hierarchy = [
                ResultOutcome.SKIP,
                ResultOutcome.PASS,
                ResultOutcome.WARN,
                ResultOutcome.FAIL,
            ]

            outcome_indices = [hierarchy.index(outcome) for outcome in outcomes]
            actual_outcome = original_outcome = hierarchy[max(outcome_indices)]

        # Find a usable log - the first one matching our "interim" outcome.
        # We cannot use the "actual" outcome, because that one may not even
        # exist in the results file - tmt might have conjured it based on
        # provided results, or set it to "error" because of errors. Only
        # the "interim" is guaranteed to be found among the results.
        test_logs = [invocation.relative_path / TEST_OUTPUT_FILENAME]

        if original_outcome is not None:
            for result in results:
                if result.result != original_outcome.value:
                    continue

                if result.log:
                    test_logs.append(invocation.relative_test_data_path / result.log[0])

                break

        return [
            tmt.Result.from_test_invocation(
                invocation=invocation,
                result=actual_outcome,
                log=test_logs,
                note=[note] if note else [],
                subresult=[result.to_subresult() for result in results],
            )
        ]

    @classmethod
    def extract_results(
        cls,
        invocation: 'TestInvocation',
        results: list[tmt.result.Result],
        logger: tmt.log.Logger,
    ) -> list[tmt.result.Result]:
        """
        Check result of a shell test.

        If there are no extra results (e.g. extracted from the tmt-report-results.yaml), continue
        normally - set the main result outcome according to test exit status.

        Otherwise, process given results, reduce their outcomes into a single one and set these
        results as tmt subresults.

        :param invocation: test invocation to which the results belong to.
        :param results: results to reduce and save as tmt subresults.
        :returns: list of results.
        """
        assert invocation.return_code is not None
        note: list[str] = []

        # Handle the `tmt-report-result` command results as a single test with assigned tmt
        # subresults.
        if results:
            return cls._process_results_reduce(invocation, results)

        # If no extra results were passed (e.g. `tmt-report-result` was not called during the
        # test), just process the exit code of a shell test and return the result.
        try:
            # Process the exit code and prepare the log path
            result = {0: ResultOutcome.PASS, 1: ResultOutcome.FAIL}[invocation.return_code]
        except KeyError:
            result = ResultOutcome.ERROR
            # Add note about the exceeded duration
            if invocation.return_code == tmt.utils.ProcessExitCodes.TIMEOUT:
                note.append('timeout')
                invocation.phase.timeout_hint(invocation)

            elif tmt.utils.ProcessExitCodes.is_pidfile(invocation.return_code):
                note.append('pidfile locking')

        return [
            tmt.Result.from_test_invocation(
                invocation=invocation,
                result=result,
                log=[invocation.relative_path / tmt.steps.execute.TEST_OUTPUT_FILENAME],
                note=note,
            )
        ]
